#! /usr/bin/env python
"""CoLRev load operation: Load records from search sources into references.bib."""
from __future__ import annotations

import itertools
import string
from pathlib import Path

import colrev.exceptions as colrev_exceptions
import colrev.loader.load_utils_formatter
import colrev.process.operation
import colrev.record.record
import colrev.settings
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import OperationsType
from colrev.constants import RecordState
from colrev.constants import SearchType


class Load(colrev.process.operation.Operation):
    """Load the records"""

    type = OperationsType.load

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
        hide_load_explanation: bool = False,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.quality_model = review_manager.get_qm()
        self.package_manager = self.review_manager.get_package_manager()

        self.load_formatter = colrev.loader.load_utils_formatter.LoadFormatter()

        if not hide_load_explanation:
            self.review_manager.logger.info("Load")
            self.review_manager.logger.info(
                "Load converts search results and adds them to the shared "
                f"{self.review_manager.paths.records}."
            )
            self.review_manager.logger.info(
                "Original records (search results) are stored in the directory data/search"
            )
            self.review_manager.logger.info(
                "See https://colrev-environment.github.io/"
                "colrev/manual/metadata_retrieval/load.html"
            )

    def _get_currently_imported_origin_list(self) -> list:
        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())
        imported_origins = [
            item for x in record_header_list for item in x[Fields.ORIGIN]
        ]
        return imported_origins

    def ensure_append_only(self, file: Path) -> None:
        """Ensure that the file was only appended to.

        This method must be called for all packages that work
        with an ex-post assignment of incremental IDs."""

        git_repo = self.review_manager.dataset.get_repo()

        # Ensure the path uses forward slashes, which is compatible with Git's path handling
        search_file_path = str(Path("data/search") / file.name).replace("\\", "/")
        revlist = (
            (
                commit.hexsha,
                (commit.tree / search_file_path).data_stream.read(),
            )
            for commit in git_repo.iter_commits(paths=str(file))
        )
        prior_file_content = ""
        for commit, filecontents in list(revlist):
            if not filecontents.decode("utf-8").startswith(prior_file_content):
                raise colrev_exceptions.AppendOnlyViolation(
                    f"{file} was changed (commit: {commit})"
                )
            prior_file_content = filecontents.decode("utf-8").replace("\r", "")
        current_contents = file.read_text(encoding="utf-8").replace("\r", "")

        if not current_contents.startswith(prior_file_content):
            raise colrev_exceptions.AppendOnlyViolation(
                f"{file} was changed (uncommitted file)"
            )

    def _import_provenance(
        self,
        record: colrev.record.record.Record,
    ) -> None:
        """Set the provenance for an imported record"""

        def set_initial_import_provenance(record: colrev.record.record.Record) -> None:
            # Initialize Fields.MD_PROV
            colrev_masterdata_provenance, colrev_data_provenance = {}, {}

            for key in sorted(record.data.keys()):
                if key in FieldSet.IDENTIFYING_FIELD_KEYS:
                    if key not in colrev_masterdata_provenance:
                        colrev_masterdata_provenance[key] = {
                            "source": record.data[Fields.ORIGIN][0],
                            "note": "",
                        }
                elif key not in FieldSet.PROVENANCE_KEYS and key not in [
                    "colrev_source_identifier",
                    Fields.ID,
                    Fields.ENTRYTYPE,
                ]:
                    colrev_data_provenance[key] = {
                        "source": record.data[Fields.ORIGIN][0],
                        "note": "",
                    }

            record.data[Fields.D_PROV] = colrev_data_provenance
            record.data[Fields.MD_PROV] = colrev_masterdata_provenance

        if not record.masterdata_is_curated():
            set_initial_import_provenance(record)
            record.run_quality_model(self.quality_model)

    def _import_record(self, *, record_dict: dict) -> dict:
        self.review_manager.logger.debug(f"import_record {record_dict[Fields.ID]}: ")

        record = colrev.record.record.Record(record_dict)

        # For better readability of the git diff:
        self.load_formatter.run(record)

        self._import_provenance(record)

        if record.data[Fields.STATUS] in [
            RecordState.md_retrieved,
            RecordState.md_needs_manual_preparation,
        ]:
            record.set_status(RecordState.md_imported)

        if record.is_retracted():
            self.review_manager.logger.info(
                f"{Colors.GREEN}Found paper retract: "
                f"{record.data['ID']}{Colors.END}"
            )

        return record.get_data()

    def _validate_source_records(
        self,
        source_records_list: list,
        *,
        source: colrev.package_manager.interfaces.SearchSourceInterface,
    ) -> None:
        # pylint: disable=too-many-branches
        if len(source_records_list) == 0:
            raise colrev_exceptions.NoRecordsToImport(
                msg=f"{source} has no records to load"
            )
        for source_record in source_records_list:
            if not all(x in source_record for x in [Fields.ID, Fields.ENTRYTYPE]):
                raise colrev_exceptions.ImportException(
                    f"Record should have keys {Fields.ID} and {Fields.ENTRYTYPE}"
                )
            if any(" " in x for x in source_record.keys()):
                raise colrev_exceptions.ImportException(
                    f"Keys should not contain spaces ({source_record.keys()})"
                )
            if not all(
                x.islower()
                for x in source_record.keys()
                if x not in [Fields.ID, Fields.ENTRYTYPE, Fields.CURATION_ID]
            ):
                raise colrev_exceptions.ImportException(
                    f"Keys should be lower case ({source_record.keys()})"
                )
            if any(x == "" for x in source_record.values()):
                raise colrev_exceptions.ImportException(
                    f"Values should not be empty ({source_record})"
                )

            for key in list(source_record.keys()):
                if "." in key:  # namespaced key
                    continue
                if key not in FieldSet.STANDARDIZED_FIELD_KEYS:
                    print(
                        f"{Colors.RED}Drop non-standardized field without namespace "
                        f"({key}){Colors.END}"
                    )
                    source_record.pop(key)

            for key in FieldSet.PROVENANCE_KEYS + [
                Fields.SCREENING_CRITERIA,
            ]:
                if key in [Fields.STATUS, Fields.PDF_ID]:
                    continue
                if (
                    key == Fields.MD_PROV
                    and Fields.MD_PROV in source_record
                    and colrev.record.record.Record(
                        source_record
                    ).masterdata_is_curated()
                ):
                    continue
                if key in source_record:
                    raise colrev_exceptions.ImportException(
                        f"Key {key} should not be in imported record"
                    )

    def setup_source_for_load(
        self,
        source: colrev.package_manager.interfaces.SearchSourceInterface,
        *,
        select_new_records: bool = True,
    ) -> None:
        """
        Prepares a search source for loading records into the review manager's dataset.

        This method initializes the loading process by selecting new records from the source
        based on the `select_new_records` flag. It then prepares the source records for import
        by filtering out already imported records if `select_new_records` is True.

        Args:
            source: The search source package endpoint interface to prepare for loading.
            select_new_records: A boolean flag indicating whether to filter out records
                                that have already been imported. Defaults to True.
        """
        source_records_list = list(source.load(self).values())  # type: ignore
        self._validate_source_records(source_records_list, source=source)

        origin_prefix = source.search_source.get_origin_prefix()
        for source_record in source_records_list:
            record_origin = f"{origin_prefix}/{source_record['ID']}"
            source_record.update(colrev_origin=[record_origin])
            colrev.record.record.Record(source_record).set_status(
                target_state=RecordState.md_retrieved
            )

        imported_origins = []
        if select_new_records:
            imported_origins = self._get_currently_imported_origin_list()
            source_records_list = [
                x
                for x in source_records_list
                if x[Fields.ORIGIN][0] not in imported_origins
            ]

        source.search_source.setup_for_load(
            source_records_list=source_records_list, imported_origins=imported_origins
        )

    def load_source_records(
        self,
        source: colrev.package_manager.interfaces.SearchSourceInterface,
        *,
        keep_ids: bool,
    ) -> None:
        """
        Loads records from a specified source into the review manager's dataset.

        This method prepares the source for loading by calling `setup_source_for_load`
        and then proceeds to load the records. It takes into account whether the IDs
        of the records should be kept as is or generated anew.

        Args:
            source: The search source package endpoint interface from which records are loaded.
            keep_ids: A boolean flag indicating whether to keep the original IDs of the records.
        """
        self.review_manager.logger.debug(
            f"Load source records {source.search_source.filename}"
        )

        self.setup_source_for_load(source)
        records = self.review_manager.dataset.load_records_dict()

        self.review_manager.logger.debug(
            f"Import individual source records {source.search_source.filename}"
        )
        for source_record in source.search_source.source_records_list:
            source_record = self._import_record(record_dict=source_record)

            # Make sure not to replace existing records
            order = 0
            letters = list(string.ascii_lowercase)
            next_unique_id = source_record[Fields.ID]
            appends: list = []
            while next_unique_id in records:
                if len(appends) == 0:
                    order += 1
                    appends = list(itertools.product(letters, repeat=order))
                next_unique_id = source_record[Fields.ID] + "".join(
                    list(appends.pop(0))
                )
            source_record[Fields.ID] = next_unique_id

            records[source_record[Fields.ID]] = source_record

            self.review_manager.logger.info(
                f" {Colors.GREEN}{source_record['ID']}".ljust(46)
                + f"md_retrieved →  {source_record['colrev_status']}{Colors.END}"
            )

        self.review_manager.logger.debug(
            f"Save records {source.search_source.filename}"
        )
        self.review_manager.dataset.save_records_dict(records)
        self._validate_load(source=source)

        if source.search_source.to_import == 0:
            self.review_manager.logger.info("No additional records loaded")
            return

        if not keep_ids:
            # Set IDs based on local_index
            # (the same records are more likely to have the same ID on the same machine)
            self.review_manager.logger.debug("Set IDs")
            records = self.review_manager.dataset.set_ids()

        self.review_manager.logger.info(
            "New records loaded".ljust(38) + f"{source.search_source.to_import} records"
        )
        self.review_manager.dataset.add_setting_changes()
        self.review_manager.dataset.add_changes(source.search_source.filename)

    def _add_source_to_settings(
        self,
        source: colrev.package_manager.interfaces.SearchSourceInterface,
    ) -> None:

        # Add to settings (if new filename)
        if source.search_source.filename in [
            s.filename for s in self.review_manager.settings.sources
        ]:
            return

        self.review_manager.logger.debug(
            f"Add source to settings {source.search_source.filename}"
        )
        git_repo = self.review_manager.dataset.get_repo()
        self.review_manager.settings.sources.append(source.search_source)
        self.review_manager.save_settings()
        # Add files that were renamed (removed)
        for obj in git_repo.index.diff(None).iter_change_type("D"):
            if source.search_source.filename.stem in obj.b_path:
                self.review_manager.dataset.add_changes(Path(obj.b_path), remove=True)

    def load_active_sources(self, *, include_md: bool = False) -> list:
        """
        Loads and returns a list of active source endpoints from the settings.

        Returns:
            list: A list of active source endpoint objects.
        """
        checker = self.review_manager.get_checker()
        checker.check_sources()
        sources_settings = []
        for source in self.review_manager.settings.sources:
            assert isinstance(source, colrev.settings.SearchSource)
            sources_settings.append(source)
        sources = []
        for source in sources_settings:
            try:
                search_source_class = self.package_manager.get_package_endpoint_class(
                    package_type=EndpointType.search_source,
                    package_identifier=source.endpoint,
                )
                endpoint = search_source_class(
                    source_operation=self,
                    settings=source.model_dump(),
                )

                s_type = endpoint.search_source.search_type  # type: ignore
                if s_type == SearchType.MD and not include_md:
                    continue
                sources.append(endpoint)

            except colrev_exceptions.MissingDependencyError as exc:
                self.review_manager.logger.error(exc)
                self.review_manager.logger.error(
                    f"Cannot load records for {source.filename}"
                )
                print()

        return sources

    def _validate_load(
        self,
        *,
        source: colrev.package_manager.interfaces.SearchSourceInterface,
    ) -> None:
        imported_origins = self._get_currently_imported_origin_list()
        imported = len(imported_origins) - source.search_source.len_before

        if imported == source.search_source.to_import:
            return
        # Note : for diagnostics, it is easier if we complete the process
        # and create the commit (instead of raising an exception)
        self.review_manager.logger.error(
            f"len_before: {source.search_source.len_before}"
        )
        self.review_manager.logger.error(f"len_after: {len(imported_origins)}")

        origins_to_import = [
            o[Fields.ORIGIN] for o in source.search_source.source_records_list
        ]
        if source.search_source.to_import - imported > 0:
            self.review_manager.logger.error(
                f"{Colors.RED}PROBLEM: delta: "
                f"{source.search_source.to_import - imported} records missing{Colors.END}"
            )

            missing_origins = [
                o for o in origins_to_import if o not in imported_origins
            ]
            self.review_manager.logger.error(
                f"{Colors.RED}Records not yet imported: {missing_origins}{Colors.END}"
            )
        else:
            self.review_manager.logger.error(
                f"{Colors.RED}PROBLEM: "
                f"{-1*(source.search_source.to_import - imported)}"
                f" records too much{Colors.END}"
            )
            additional_origins = [
                o for o in imported_origins if o not in origins_to_import
            ]
            self.review_manager.logger.error(
                f"{Colors.RED}Records additionally imported: {additional_origins}{Colors.END}"
            )

    def _create_load_commit(self, source: colrev.settings.SearchSource) -> None:
        git_repo = self.review_manager.dataset.get_repo()
        stashed = "No local changes to save" != git_repo.git.stash(
            "push", "--keep-index"
        )
        # part_exact_call = self.review_manager.exact_call
        # self.review_manager.exact_call = (
        #     f"{part_exact_call} -s {source.search_source.filename.name}"
        # )
        self.review_manager.dataset.create_commit(
            msg=f"Load: data/search/{source.search_source.filename.name} → "
            f"{self.review_manager.paths.RECORDS_FILE_GIT}",
            skip_hooks=True,
        )
        if stashed:
            git_repo.git.stash("pop")
        if not self.review_manager.high_level_operation:
            print()

    @colrev.process.operation.Operation.decorate()
    def main(
        self,
        *,
        keep_ids: bool = False,
    ) -> None:
        """Load records (main entrypoint)"""

        if not self.review_manager.high_level_operation:
            print()

        for source in self.load_active_sources():
            try:
                self.review_manager.logger.info(f"Load {source.search_source.filename}")
                self._add_source_to_settings(source)
                self.load_source_records(source, keep_ids=keep_ids)
                self._create_load_commit(source)

            except FileNotFoundError:
                self.review_manager.logger.info(" Nothing to load")
                print()
            except colrev_exceptions.NoRecordsToImport as exc:
                self.review_manager.logger.info(exc)
                print()
            except colrev_exceptions.ImportException as exc:
                self.review_manager.logger.error(f"{Colors.RED}{exc}{Colors.END}")

        self.review_manager.logger.info(
            f"{Colors.GREEN}Completed load operation{Colors.END}"
        )



#! /usr/bin/env python
"""SearchSource: OpenCitations"""
from __future__ import annotations

import json
import typing
from pathlib import Path

import requests
import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import RecordState
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType
from colrev.packages.crossref.src import crossref_api

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class OpenCitationsSearchSource:
    """Forward search based on OpenCitations
    Scope: all included papers with colrev_status in (rev_included, rev_synthesized)
    """

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings
    endpoint = "colrev.open_citations_forward_search"
    source_identifier = "fwsearch_ref"
    search_types = [SearchType.FORWARD_SEARCH]

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.supported

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.review_manager = source_operation.review_manager
        self.crossref_api = crossref_api.CrossrefAPI(params={})

    @classmethod
    def get_default_source(cls) -> colrev.settings.SearchSource:
        """Get the default SearchSource settings"""
        return colrev.settings.SearchSource(
            endpoint="colrev.open_citations_forward_search",
            filename=Path("data/search/forward_search.bib"),
            search_type=SearchType.FORWARD_SEARCH,
            search_parameters={
                "scope": {Fields.STATUS: "rev_included|rev_synthesized"}
            },
            comment="",
        )

    def _validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""

        source = self.search_source

        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        assert source.search_type == SearchType.FORWARD_SEARCH

        if "scope" not in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "Scope required in the search_parameters"
            )

        if (
            source.search_parameters["scope"][Fields.STATUS]
            != "rev_included|rev_synthesized"
        ):
            raise colrev_exceptions.InvalidQueryException(
                "search_parameters/scope/colrev_status must be rev_included|rev_synthesized"
            )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    def _fw_search_condition(self, *, record: dict) -> bool:
        if Fields.DOI not in record:
            return False

        # rev_included/rev_synthesized required, but record not in rev_included/rev_synthesized
        if (
            Fields.STATUS in self.search_source.search_parameters["scope"]
            and self.search_source.search_parameters["scope"][Fields.STATUS]
            == "rev_included|rev_synthesized"
            and record[Fields.STATUS]
            not in [
                RecordState.rev_included,
                RecordState.rev_synthesized,
            ]
        ):
            return False

        return True

    def _get_forward_search_records(self, *, record_dict: dict) -> list:
        forward_citations = []

        url = f"https://opencitations.net/index/coci/api/v1/citations/{record_dict['doi']}"
        # headers = {"authorization": "YOUR-OPENCITATIONS-ACCESS-TOKEN"}
        headers: typing.Dict[str, str] = {}

        ret = requests.get(url, headers=headers, timeout=300)
        try:
            items = json.loads(ret.text)

            for doi in [x["citing"] for x in items]:
                retrieved_record = self.crossref_api.query_doi(doi=doi)
                # if not crossref_query_return:
                #     raise colrev_exceptions.RecordNotFoundInPrepSourceException()
                retrieved_record.data[Fields.ID] = retrieved_record.data[Fields.DOI]
                forward_citations.append(retrieved_record.data)
        except json.decoder.JSONDecodeError:
            self.review_manager.logger.info(
                f"Error retrieving citations from Opencitations for {record_dict['ID']}"
            )

        return forward_citations

    def search(self, rerun: bool) -> None:
        """Run a forward search based on OpenCitations"""

        # pylint: disable=too-many-branches

        self._validate_source()

        records = self.review_manager.dataset.load_records_dict()

        if not records:
            print("No records imported. Cannot run forward search yet.")
            return

        forward_search_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        for record in records.values():
            if not self._fw_search_condition(record=record):
                continue

            self.review_manager.logger.info(
                f"Run forward search for {record[Fields.ID]}"
            )

            new_records = self._get_forward_search_records(record_dict=record)

            for new_record in new_records:
                try:
                    new_record["fwsearch_ref"] = (
                        record[Fields.ID] + "_forward_search_" + new_record[Fields.ID]
                    )
                    new_record["cites_IDs"] = record[Fields.ID]
                    retrieved_record = colrev.record.record.Record(new_record)

                    forward_search_feed.add_update_record(
                        retrieved_record=retrieved_record
                    )
                except colrev_exceptions.NotFeedIdentifiableException:
                    continue

        forward_search_feed.save()

        if self.review_manager.dataset.has_record_changes():
            self.review_manager.dataset.create_commit(
                msg="Forward search", script_call="colrev search"
            )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for forward searches (OpenCitations)"""

        result = {"confidence": 0.0}

        return result

    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint"""

        search_source = cls.get_default_source()
        operation.add_source_and_search(search_source)
        return search_source

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for forward searches (OpenCitations)"""
        return record



#! /usr/bin/env python
"""Convenience functions to load tabular files (csv, xlsx)"""
from __future__ import annotations

import logging
import typing
from pathlib import Path

import pandas as pd

import colrev.exceptions as colrev_exceptions
import colrev.loader.loader


# pylint: disable=too-few-public-methods
# pylint: disable=too-many-arguments


class TableLoader(colrev.loader.loader.Loader):
    """Loads csv and Excel files (based on pandas)"""

    def __init__(
        self,
        *,
        filename: Path,
        entrytype_setter: typing.Callable,
        field_mapper: typing.Callable,
        id_labeler: typing.Callable,
        unique_id_field: str = "",
        logger: logging.Logger = logging.getLogger(__name__),
    ):
        super().__init__(
            filename=filename,
            id_labeler=id_labeler,
            unique_id_field=unique_id_field,
            entrytype_setter=entrytype_setter,
            field_mapper=field_mapper,
            logger=logger,
        )

    @classmethod
    def get_nr_records(cls, filename: Path) -> int:
        """Get the number of records in the file"""
        if filename.name.endswith(".csv"):
            data = pd.read_csv(filename)
        elif filename.name.endswith((".xls", ".xlsx")):
            data = pd.read_excel(filename, dtype=str)
        else:
            raise NotImplementedError
        count = len(data)
        return count

    def load_records_list(self) -> list:
        try:
            if self.filename.name.endswith(".csv"):
                data = pd.read_csv(self.filename)
            elif self.filename.name.endswith((".xls", ".xlsx")):
                data = pd.read_excel(
                    self.filename, dtype=str
                )  # dtype=str to avoid type casting
            else:
                raise NotImplementedError

        except pd.errors.ParserError as exc:  # pragma: no cover
            raise colrev_exceptions.ImportException(
                f"Error: Not a valid file? {self.filename.name}"
            ) from exc

        records_list = data.to_dict("records")
        return records_list



#! /usr/bin/env python
"""Screen based on CLI"""
from __future__ import annotations

import zope.interface
from inquirer import Checkbox
from inquirer import prompt
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.packages.screen_utils as util_cli_screen
import colrev.record.record
import colrev.settings
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import ScreenCriterionType


@zope.interface.implementer(colrev.package_manager.interfaces.ScreenInterface)
class CoLRevCLIScreen:
    """Screen documents using a CLI"""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=False)

    def __init__(
        self,
        *,
        screen_operation: colrev.ops.screen.Screen,
        settings: dict,
    ) -> None:
        self.review_manager = screen_operation.review_manager
        self.screen_operation = screen_operation
        self.settings = self.settings_class(**settings)

        self._i = 0
        self._pad = 0
        self._stat_len = 0
        self.screening_criteria: dict = {}
        self.criteria_available = 0

    def _print_screening_criteria(self) -> None:
        if not self.review_manager.settings.screen.criteria:
            return
        print("\nIn the screen, the following criteria are applied:\n")
        for (
            criterion_name,
            criterion_settings,
        ) in self.review_manager.settings.screen.criteria.items():
            color = Colors.GREEN
            if (
                ScreenCriterionType.exclusion_criterion
                == criterion_settings.criterion_type
            ):
                color = Colors.RED
            print(
                f" - {criterion_name} "
                f"({color}{criterion_settings.criterion_type}{Colors.END}): "
                f"{criterion_settings.explanation}"
            )
            if criterion_settings.comment:
                print(f"   {criterion_settings.comment}")

    def _screen_with_criteria_print_overall_decision(
        self, *, record: colrev.record.record.Record, screen_inclusion: bool
    ) -> None:
        if screen_inclusion:
            print(
                f"Overall screening decision for {record.data['ID']}: "
                f"{Colors.GREEN}include{Colors.END}"
            )
        else:
            print(
                f"Overall screening decision for {record.data['ID']}: "
                f"{Colors.RED}exclude{Colors.END}"
            )

    # pylint: disable=too-many-branches
    def _screen_record_with_criteria(
        self,
        record: colrev.record.record.Record,
    ) -> str:
        choices = []
        for criterion_name, criterion_settings in self.screening_criteria.items():
            color = Colors.GREEN
            if (
                ScreenCriterionType.exclusion_criterion
                == criterion_settings.criterion_type
            ):
                color = Colors.RED
            choices.append(f"{color}{criterion_name}{Colors.END} violated")
        choices.append("Include (no criterion violated)")
        choices.append("Skip")
        choices.append("Quit")
        question = [
            Checkbox(
                "violated_criteria",
                message="Select the criteria that are violated "
                + "[use space to select (if any), use Enter to confirm]:",
                choices=choices,
            ),
        ]

        while True:
            answers = prompt(question)
            violated_criteria = answers["violated_criteria"]

            # can only select one of the options quit, skip, or include/no criterion violated
            if "Quit" in violated_criteria and len(violated_criteria) > 1:
                print(
                    "Contradiction: 'Quit' and other criteria selected. Please select again."
                )
                continue
            if "Skip" in violated_criteria and len(violated_criteria) > 1:
                print(
                    "Contradiction: 'Skip' and other criteria selected. Please select again."
                )
                continue
            if (
                "Include (no criterion violated)" in violated_criteria
                and len(violated_criteria) > 1
            ):
                print(
                    "Contradiction: 'Include (no criterion violated)'"
                    " and other criteria selected. Please select again."
                )
                continue
            # prevent contradiction (include and criterion violated)
            if "Include (no criterion violated)" in violated_criteria:
                violated_criteria.remove("Include (no criterion violated)")
                if violated_criteria:
                    print(
                        "Contradiction: 'Include (no criterion violated)' "
                        "and other criteria selected. Please select again."
                    )
                    continue
            break

        if "Skip" in violated_criteria:
            return "skip"
        if "Quit" in violated_criteria:
            return "quit"

        decisions = {c: "in" for c in self.screening_criteria.keys()}
        if "Include (no criterion violated)" not in violated_criteria:
            for criterion_name in violated_criteria:
                decisions[
                    criterion_name.replace(Colors.RED, "")
                    .replace(Colors.END, "")
                    .replace(Colors.GREEN, "")
                    .replace(" violated", "")
                ] = "out"

        c_field = ""
        for criterion_name, decision in decisions.items():
            c_field += f";{criterion_name}={decision}"
        c_field = c_field.replace(" ", "").lstrip(";")

        screen_inclusion = all(decision == "in" for _, decision in decisions.items())

        self._screen_with_criteria_print_overall_decision(
            record=record, screen_inclusion=screen_inclusion
        )

        self.screen_operation.screen(
            record=record,
            screen_inclusion=screen_inclusion,
            screening_criteria=c_field,
            PAD=self._pad,
        )
        return "screened"

    def _screen_record_without_criteria(
        self,
        record: colrev.record.record.Record,
    ) -> str:
        quit_pressed = False
        decision, ret = "NA", "NA"
        while ret not in ["y", "n", "q", "s"]:
            ret = input(
                f"({self._i}/{self._stat_len}) "
                "Include [y,n,q,s for yes, no, quit, skip/screen later]? "
            )
            if ret == "q":
                quit_pressed = True
            elif ret == "s":
                return "skip"
            elif ret in ["y", "n"]:
                decision = ret

        if quit_pressed:
            self.review_manager.logger.info("Stop screen")
            return "quit"

        if decision == "y":
            self.screen_operation.screen(
                record=record,
                screen_inclusion=True,
                screening_criteria="NA",
            )
        if decision == "n":
            self.screen_operation.screen(
                record=record,
                screen_inclusion=False,
                screening_criteria="NA",
                PAD=self._pad,
            )
        return "screened"

    def _screen_record(
        self,
        *,
        record_dict: dict,
    ) -> str:
        record = colrev.record.record.Record(record_dict)

        self._i += 1
        print("\n\n")
        print(f"Record {self._i} (of {self._stat_len})")
        print()
        print(record.data["ID"])
        print()
        record.print_citation_format()
        if "abstract" in record.data:
            print()
            print(record.data["abstract"])
        print()
        if self.criteria_available:
            ret = self._screen_record_with_criteria(record)

        else:
            ret = self._screen_record_without_criteria(record)

        return ret

    def _screen_cli(self, split: list) -> dict:
        screen_data = self.screen_operation.get_data()
        self._pad = screen_data["PAD"]
        self._stat_len = screen_data["nr_tasks"]

        if 0 == self._stat_len:
            self.review_manager.logger.info("No records to prescreen")

        records = self.review_manager.dataset.load_records_dict()

        self._print_screening_criteria()

        self.screening_criteria = (
            util_cli_screen.get_screening_criteria_from_user_input(
                screen_operation=self.screen_operation, records=records
            )
        )
        self.criteria_available = len(self.screening_criteria.keys())

        for record_dict in screen_data["items"]:
            if record_dict[Fields.ID] not in split:
                continue

            ret = self._screen_record(record_dict=record_dict)

            if ret == "skip":
                continue
            if ret == "quit":
                self.review_manager.logger.info("Stop screen")
                break

        if self._stat_len == 0:
            self.review_manager.logger.info("No records to screen")
            return records

        if self._i < self._stat_len and split:  # if records remain for screening
            if input("Create commit (y/n)?") != "y":
                return records

        self.review_manager.dataset.create_commit(
            msg="Screen: manual (cli)", manual_author=True
        )
        return records

    def run_screen(self, records: dict, split: list) -> dict:
        """Screen records based on a cli"""

        records = self._screen_cli(split)

        return records



#!/usr/bin/env python
"""Console UI for Semantic Scholar"""
import datetime
import re
import typing

import inquirer

import colrev.exceptions as colrev_exceptions
from colrev.constants import Fields


class SemanticScholarUI:
    """Implements the User Interface for the SemanticScholar API Search within colrev"""

    search_params: dict

    def __init__(self) -> None:
        self.search_params = {}
        self.search_subject = ""

    def main_ui(self) -> None:
        """Display the main Menu and choose the search type"""

        run = True

        print("\nWelcome to SemanticScholar! \n\n")
        while run:
            main_msg = "Please choose one of the options below "
            main_options = [
                "Keyword search",
                "Search for paper by ID",
                "Search for author by ID",
                "Exit Program",
            ]

            fwd_value = self.choose_single_option(msg=main_msg, options=main_options)

            if fwd_value == "Search for paper by ID":
                self.search_subject = "paper"
                run = self.paper_ui()

            elif fwd_value == "Search for author by ID":
                self.search_subject = "author"
                run = self.author_ui()

            elif fwd_value == "Keyword search":
                self.search_subject = "keyword"
                self.keyword_ui()
                run = False

            elif fwd_value == "Exit Program":
                print("\nThanks for using Colrev! This Program will close.")
                raise colrev_exceptions.ServiceNotAvailableException("Regular exit.")

        if not self.search_params:
            print("\n Search cancelled. This program will close.")
            raise colrev_exceptions.ServiceNotAvailableException(
                "No search parameters were entered."
            )

    def paper_ui(self) -> bool:
        """Ask user to enter search parameters for distinctive paper search"""

        paper_id_list = []

        while True:
            p_msg = "How would you like to search for the paper?"
            p_options = [
                "S2PaperId",
                "CorpusId",
                "DOI",
                "ArXivId",
                "MAG",
                "ACL",
                "PMID",
                "PMCID",
            ]

            param = self.choose_single_option(msg=p_msg, options=p_options)

            if param in p_options:
                id_formats = ["S2PaperId", "DOI", "ArXivId", "ACL"]
                param_value = self.enter_text(
                    msg="Please enter the chosen ID in the right format "
                )

                if param in id_formats:
                    correct_format = self.check_format(param=param, value=param_value)

                    while not correct_format and param_value:
                        param_value = self.enter_text(
                            msg="Error: Invalid ID format. Please try again or press Enter."
                        )

                        correct_format = self.check_format(
                            param=param, value=param_value
                        )

                else:
                    while (
                        not self.id_validation_with_regex(
                            id_value=param_value, regex=r"^[0-9]+$"
                        )
                        and param_value
                    ):
                        param_value = self.enter_text(
                            msg="Error: Invalid ID format. Please try again or press Enter."
                        )

                if param_value:
                    paper_id_list.append(param_value)
                    self.search_params["paper_ids"] = paper_id_list

            fwd = self.choose_single_option(
                msg="How would you like to continue?",
                options=[
                    "Conduct Search",
                    "Search for another paper or enter different ID",
                    "Back to main Menu",
                ],
            )

            if fwd == "Conduct Search":
                return False

            if fwd == "Back to main Menu":
                return True

    def author_ui(self) -> bool:
        """Ask user to enter search parameters for distinctive author search"""

        author_id_list = []

        while True:
            validation_break = False

            param_value = self.enter_text(
                msg="Please enter an S2 author ID in the right format "
            )
            while (
                not self.id_validation_with_regex(
                    id_value=param_value, regex=r"^[a-zA-Z0-9]+$"
                )
                and not validation_break
            ):
                param_value = self.enter_text(
                    msg="Error: Invalid S2AuthorId format. Please try again or press Enter."
                )
                if not param_value:
                    validation_break = True

            if not validation_break:
                author_id_list.append(param_value)
                self.search_params["author_ids"] = author_id_list

            fwd = self.choose_single_option(
                msg="How would you like to continue?",
                options=[
                    "Conduct Search",
                    "Search for another author or enter different ID",
                    "Back to main Menu",
                ],
            )

            if fwd == "Conduct Search":
                return False

            if fwd == "Back to main Menu":
                return True

    def keyword_ui(self) -> None:
        """Ask user to enter Searchstring and limitations for Keyword search"""

        query = self.enter_text(msg="Please enter the query for your keyword search ")
        while not (query and isinstance(query, str)):
            query = self.enter_text(
                msg="Error: You must enter a query to conduct a search. Please enter a query"
            )

        self.search_params["query"] = query

        year = self.enter_year()
        if year:
            self.search_params[Fields.YEAR] = year

        venue = self.enter_text(
            msg="To search for papers from specific venues, enter the venues here."
            + " Separate multiple venues by comma."
            + " Please press Enter to not specify any venues "
        )
        if venue:
            self.search_params["venue"] = venue.split(",")

        fields_of_study = self.enter_study_fields()
        if fields_of_study:
            self.search_params["fields_of_study"] = fields_of_study

        open_access = self.choose_single_option(
            msg="Would you like to only search for items "
            "for which the full text is available as pdf?",
            options=["NO", "YES"],
        )
        if open_access == "YES":
            self.search_params["open_access_pdf"] = True
        else:
            self.search_params["open_access_pdf"] = False

    def get_api_key(self, existing_key: typing.Optional[str] = "") -> str:
        """Method to get API key from user input"""

        ask_again = True

        if existing_key:
            api_key = existing_key
        else:
            print("\n")
            api_key = self.enter_text(
                msg="Please enter a valid API key for SemanticScholar. "
                "If you don't have a key, please press Enter."
            )

        while ask_again:
            ask_again = False

            if not api_key:
                print(
                    "\nWARNING: Searching without an API key might not be successful. \n"
                )
                fwd = self.choose_single_option(
                    msg="Would you like to continue?", options=["YES", "NO"]
                )

                if fwd == "NO":
                    api_key = self.enter_text(msg="Please enter an API key ")
                    ask_again = True
                else:
                    return ""

            elif not re.match(r"^\w{40}$", api_key):
                print("Error: Invalid API key.\n")
                fwd = self.choose_single_option(
                    msg="Would you like to enter a different key?",
                    options=["YES", "NO"],
                )

                if fwd == "YES":
                    api_key = self.enter_text(msg="Please enter an API key ")
                    ask_again = True
                else:
                    return ""

            else:
                print("\n" + "API key: " + api_key + "\n")
                fwd = self.choose_single_option(
                    msg="Start search with this API key?", options=["YES", "NO"]
                )

                if fwd == "NO":
                    api_key = self.enter_text(msg="Please enter a different API key ")
                    ask_again = True

        return api_key

    def enter_year(self) -> str:
        """Method to ask a specific year span in the format allowed by the SemanticScholar API"""

        examples = (
            "Examples for valid year spans: '2019'; '2012-2020'; '-2022'; '2015-'"
        )
        ask_again = True
        year_span = self.enter_text(
            msg="Please enter a year span. "
            "Please press Enter if you don't wish to specify a year span"
        )
        while year_span and ask_again:
            ask_again = False
            if not re.match(
                "|".join([r"^-\d{4}$", r"^\d{4}-?$", r"^\d{4}-\d{4}$"]), year_span
            ):
                print("Error: Invalid year span.\n" + examples + "\n")
                year_span = self.enter_text(
                    msg="Please enter a year span."
                    + " Please press Enter if you don't wish to specify a year span"
                )
                ask_again = True
            elif re.match(r"^\d{4}-\d{4}$", year_span):
                years = year_span.split("-")
                year_a = int(years[0])
                year_b = int(years[1])
                if year_a >= year_b or (year_b > int(datetime.date.today().year)):
                    print("Error: Invalid year span.\n" + examples + "\n")
                    year_span = self.enter_text(
                        msg="Please enter a year span."
                        + " Please press Enter if you don't wish to specify a year span"
                    )
                    ask_again = True
            elif re.match(r"^-?\d{4}-?$", year_span):
                year = int(re.findall(r"\d{4}", year_span)[0])
                if year > int(datetime.date.today().year):
                    print(
                        "Error: Invalid year span. You cannot search for papers from the future.\n"
                    )
                    year_span = self.enter_text(
                        msg="Please enter a year span."
                        + " Please press Enter if you don't wish to specify a year span"
                    )
                    ask_again = True

        return year_span

    def enter_study_fields(self) -> list:
        """Method to ask a selection of fields of
        study that are allowed by the Semantic Scholar API"""

        msg = (
            "If you want to restrict your search to certain study fields, "
            "select them here or press Enter"
        )
        options = [
            "Computer Science",
            "Medicine",
            "Chemistry",
            "Biology",
            "Materials Science",
            "Physics",
            "Geology",
            "Psychology",
            "Art",
            "History",
            "Geography",
            "Sociology",
            "Business",
            "Political Science",
            "Economics",
            "Philosophy",
            "Mathematics",
            "Engineering",
            "Environmental Science",
            "Agricultural and Food Sciences",
            "Education",
            "Law",
            "Linguistics",
        ]
        study_fields = self.choose_multiple_options(msg=msg, options=options)

        return study_fields

    def choose_single_option(
        self,
        *,
        msg: str,
        options: list,
    ) -> str:
        """Method to display a question with single choice answers to the console using inquirer"""

        question = [
            inquirer.List(
                name="Choice",
                message=msg,
                choices=options,
                carousel=False,
            ),
        ]
        choice = inquirer.prompt(questions=question)

        return choice.get("Choice")

    def choose_multiple_options(
        self,
        *,
        msg: str,
        options: list,
    ) -> list:
        """Method to display a question with multiple
        choice answers to the console using inquirer"""

        question = [
            inquirer.Checkbox(
                name="Choice",
                message=msg,
                choices=options,
                carousel=False,
            ),
        ]
        choice = inquirer.prompt(questions=question)

        return choice.get("Choice")

    def enter_text(
        self,
        *,
        msg: str,
    ) -> str:
        """Method to display a question with free text
        entry answer to the console using inquirer."""

        question = [
            inquirer.Text(
                name="Entry",
                message=msg,
            )
        ]
        choice = inquirer.prompt(questions=question)

        return choice.get("Entry")

    def id_validation_with_regex(
        self,
        *,
        id_value: str,
        regex: str,
    ) -> bool:
        """Method to validate ID formats using a regex as an argument"""

        if re.match(regex, id_value):
            return True

        return False

    def check_format(self, *, param: str, value: str) -> bool:
        """Method to validate certain ID formats"""
        if param == "S2PaperId":
            if self.id_validation_with_regex(id_value=value, regex=r"^[a-zA-Z0-9]+$"):
                return True
        elif param == "DOI":
            if self.id_validation_with_regex(id_value=value, regex=r"^10\..+$"):
                return True

        elif param == "ArXivId":
            if self.id_validation_with_regex(id_value=value, regex=r"^\d+\.\d+$"):
                return True

        elif param == "ACL":
            if self.id_validation_with_regex(id_value=value, regex=r"^\w+-\w+$"):
                return True

        return False



#! /usr/bin/env python
"""CLI interface for manual retrieval of PDFs"""
from __future__ import annotations

import shutil
import urllib.parse
from pathlib import Path

import zope.interface
from pydantic import Field

import colrev.env.utils
import colrev.ops.pdf_get
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import Fields
from colrev.constants import RecordState


# pylint: disable=too-few-public-methods
# pylint: disable=unused-argument


@zope.interface.implementer(colrev.package_manager.interfaces.PDFGetManInterface)
class CoLRevCLIPDFGetMan:
    """Get PDFs manually based on a CLI"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings
    ci_supported: bool = Field(default=False)

    def __init__(
        self,
        *,
        pdf_get_man_operation: colrev.ops.pdf_get_man.PDFGetMan,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.review_manager = pdf_get_man_operation.review_manager
        self.pdf_get_man_operation = pdf_get_man_operation

        self._get_from_downloads_folder = False
        self.pdf_dir = self.review_manager.paths.pdf

    def _get_pdf_from_google(
        self, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        # import webbrowser

        title = record.data.get(Fields.TITLE, "no title")
        title = urllib.parse.quote_plus(title)
        url = f"  google:   https://www.google.com/search?q={title}+filetype%3Apdf"
        # webbrowser.open_new_tab(url)
        print(url)
        return record

    def _ask_authors_for_pdf(
        self, record: colrev.record.record.Record
    ) -> colrev.record.record.Record:
        # get the recipient email(s) from the local author index
        recipient = "TODO"
        subject = f"Copy of a PDF ({record.data['ID']})"

        author_name = record.data.get(Fields.AUTHOR, "").split(",")[0]
        signed, _ = self.review_manager.get_committer()

        template = colrev.env.utils.get_template(
            "packages/colrev_cli_pdf_get_man/src/pdf_get_man_mail.txt"
        )

        content = template.render(record=record, author_name=author_name, signed=signed)

        print("\n\n")
        print(recipient)
        print(subject)
        print("\n\n")
        print(content)
        print("\n\n")

        # next steps: better integration with email clients

        return record

    def _get_filepath(self, *, record: colrev.record.record.Record) -> Path:
        filepath = (
            self.pdf_dir / f"{record.data.get('year', 'NA')}/{record.data['ID']}.pdf"
        )
        if filepath.is_file():
            return filepath

        if Fields.VOLUME in record.data:
            filepath = self.pdf_dir / f"{record.data['volume']}/{record.data['ID']}.pdf"
            if filepath.is_file():
                return filepath

        if Fields.VOLUME in record.data and Fields.NUMBER in record.data:
            filepath = (
                self.pdf_dir
                / f"{record.data['volume']}_{record.data['number']}/{record.data['ID']}.pdf"
            )
            if filepath.is_file():
                return filepath

            filepath = (
                self.pdf_dir
                / f"{record.data['volume']}/{record.data['number']}/{record.data['ID']}.pdf"
            )
            if filepath.is_file():
                return filepath

        filepath = self.pdf_dir / f"{record.data['ID']}.pdf"

        return filepath

    def _retrieve_record_from_downloads_folder(
        self,
        *,
        record: colrev.record.record.Record,
    ) -> None:
        downloads_folder = Path.home() / Path("Downloads")
        pdfs_in_downloads_folder = list(downloads_folder.glob("*.pdf"))

        if len(pdfs_in_downloads_folder) == 0:
            print("No PDF found in downloads_folder")
            return
        if (
            len(pdfs_in_downloads_folder) > 1
            and not (downloads_folder / Path(f"{record.data['ID']}.pdf")).is_file()
        ):
            print("Multiple PDFs found in downloads_folder (skipping)")
            return

        if Path(f"{record.data['ID']}.pdf").is_file():
            pdf_in_downloads_folder = downloads_folder / Path(
                f"{record.data['ID']}.pdf"
            )
        else:
            pdf_in_downloads_folder = pdfs_in_downloads_folder[0]

        # simple heuristics:
        vol_slash_nr_path = self.pdf_dir / Path(
            f"{record.data.get('volume', 'NA')}/{record.data.get('number', 'NA')}"
        )
        if vol_slash_nr_path.is_dir():
            shutil.move(
                str(pdf_in_downloads_folder),
                str(vol_slash_nr_path / Path(f"{record.data['ID']}.pdf")),
            )
            return

        vol_underscore_nr_path = self.pdf_dir / Path(
            f"{record.data.get('volume', 'NA')}_{record.data.get('number', 'NA')}"
        )
        if vol_underscore_nr_path.is_dir():
            shutil.move(
                str(pdf_in_downloads_folder),
                str(vol_underscore_nr_path / Path(f"{record.data['ID']}.pdf")),
            )
            return

        vol_path = self.pdf_dir / Path(f"{record.data.get('volume', 'NA')}")
        if vol_path.is_dir():
            shutil.move(
                str(pdf_in_downloads_folder),
                str(vol_path / Path(f"{record.data['ID']}.pdf")),
            )
            return

        year_path = self.pdf_dir / Path(f"{record.data.get('year', 'NA')}")
        if year_path.is_dir():
            shutil.move(
                str(pdf_in_downloads_folder),
                str(year_path / Path(f"{record.data['ID']}.pdf")),
            )
            return

        shutil.move(
            str(pdf_in_downloads_folder),
            str(self.pdf_dir / Path(f"{record.data['ID']}.pdf")),
        )

    def print_record(self, *, record_dict: dict) -> None:
        """Print the record for pdf-get-man (cli)"""

        ret_str = f"  ID:       {record_dict['ID']} ({record_dict['ENTRYTYPE']})"
        ret_str += (
            f"\n  title:    {Colors.GREEN}{record_dict.get('title', 'no title')}{Colors.END}"
            f"\n  author:   {record_dict.get('author', 'no-author')}"
        )
        if record_dict[Fields.ENTRYTYPE] == "article":
            ret_str += (
                f"\n  outlet: {record_dict.get('journal', 'no-journal')} "
                f"({record_dict.get('year', 'no-year')}) "
                f"{record_dict.get('volume', 'no-volume')}"
                f"({record_dict.get('number', '')})"
            )
        elif record_dict[Fields.ENTRYTYPE] == "inproceedings":
            ret_str += f"\n  {record_dict.get('booktitle', 'no-booktitle')}"
        if Fields.FULLTEXT in record_dict:
            ret_str += (
                f"\n  fulltext: {Colors.ORANGE}{record_dict['fulltext']}{Colors.END}"
            )

        if Fields.URL in record_dict:
            ret_str += f"\n  url:      {record_dict['url']}"

        print(ret_str)

    def _pdf_get_man_record_cli(
        self,
        *,
        record: colrev.record.record.Record,
    ) -> None:
        # self.review_manager.logger.debug(
        #     f"called pdf_get_man_cli for {record}"
        # )

        # to print only the essential information
        self.print_record(record_dict=record.get_data())

        if RecordState.pdf_needs_manual_retrieval != record.data[Fields.STATUS]:
            return

        retrieval_scripts = {
            "get_pdf_from_google": self._get_pdf_from_google,
            "ask_authors": self._ask_authors_for_pdf,
            # 'get_pdf_from_researchgate': get_pdf_from_researchgate,
        }

        filepath = self._get_filepath(record=record)
        for retrieval_script in retrieval_scripts.values():
            # self.review_manager.logger.debug(
            #     f'{script_name}({record.data[Fields.ID]}) called'
            # )
            record = retrieval_script(record)

            if input("Retrieved (y/n)?") == "y":
                if self._get_from_downloads_folder:
                    self._retrieve_record_from_downloads_folder(record=record)
                filepath = self._get_filepath(record=record)
                if not filepath.is_file():
                    print(f"File does not exist: {record.data[Fields.ID]}.pdf")
                else:
                    filepath = self._get_filepath(record=record)
                    self.pdf_get_man_operation.pdf_get_man_record(
                        record=record,
                        filepath=filepath,
                    )
                    break

        if not filepath.is_file():
            if input("Is the PDF available (y/n)?") == "n":
                self.pdf_get_man_operation.pdf_get_man_record(
                    record=record,
                    filepath=None,
                )

        self.review_manager.update_status_yaml()

    def pdf_get_man(self, records: dict) -> dict:
        """Get the PDF manually based on a cli"""

        self.review_manager.logger.info("Retrieve PDFs manually")
        pdf_get_operation = self.review_manager.get_pdf_get_operation()
        pdf_dir = self.pdf_dir

        records = self.review_manager.dataset.load_records_dict()
        if input("Check existing unlinked PDFs (y/n)?") == "y":
            records = pdf_get_operation.check_existing_unlinked_pdfs(records)

        if input("Get PDF from Downloads folder (y/n)?") == "y":
            self._get_from_downloads_folder = True

        for record_dict in records.values():
            record = colrev.record.record.Record(record_dict)
            pdf_get_operation.link_pdf(record).get_data()

        self.pdf_get_man_operation.export_retrieval_table(records)
        pdf_get_man_data = self.pdf_get_man_operation.get_data()
        if pdf_get_man_data["nr_tasks"] == 0:
            self.review_manager.logger.info(
                "No tasks for PDF retrieval (run colrev pdf-get )."
            )
            return records
        print(
            "\nInstructions\n\n      "
            "Get the pdfs, rename them (ID.pdf) and store them in the pdfs directory.\n"
        )
        input("Enter to start.")

        for i, item in enumerate(pdf_get_man_data["items"]):
            stat = str(i + 1) + "/" + str(pdf_get_man_data["nr_tasks"])

            record = colrev.record.record.Record(records[item[Fields.ID]])

            print(f"\n\n{stat}")

            self._pdf_get_man_record_cli(record=record)

        if self.review_manager.dataset.has_record_changes():
            if input("Create commit (y/n)?") == "y":
                self.review_manager.dataset.create_commit(
                    msg="Retrieve PDFs manually",
                    manual_author=True,
                )
        else:
            self.review_manager.logger.info(
                "Retrieve PDFs manually and copy the files to "
                f"the {pdf_dir}. Afterwards, use "
                "colrev pdf-get-man"
            )

        return records



#! /usr/bin/env python
"""Functionality to determine similarity betwen records."""
from __future__ import annotations

import re
import typing

import colrev.env.utils
from colrev.constants import DefectCodes
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import FieldValues

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.record.record


def _get_merging_triple(
    merging_record: colrev.record.record.Record, *, key: str, default_source: str
) -> typing.Tuple[
    str,
    str,
    str,
]:
    val = merging_record.data.get(key, "")

    # do not override provenance, ID, ... fields
    if key in [
        Fields.ID,
        Fields.MD_PROV,
        Fields.D_PROV,
        Fields.COLREV_ID,
        Fields.STATUS,
        Fields.ORIGIN,
        "MOVED_DUPE_ID",
    ]:
        val = ""

    field_provenance = merging_record.get_field_provenance(
        key=key, default_source=default_source
    )
    source = field_provenance["source"]
    note = field_provenance["note"]
    return val, source, note


def _merge_origins(
    main_record: colrev.record.record.Record,
    merging_record: colrev.record.record.Record,
) -> None:
    """Merge the origins with those of the merging_record"""

    if Fields.ORIGIN in merging_record.data:
        origins = main_record.data[Fields.ORIGIN] + merging_record.data[Fields.ORIGIN]
        main_record.data[Fields.ORIGIN] = sorted(list(set(origins)))


def _merge_status(
    main_record: colrev.record.record.Record,
    merging_record: colrev.record.record.Record,
) -> None:
    """Merge the status with the merging_record"""

    if Fields.STATUS in merging_record.data:
        # Set both status to the latter in the state model
        if main_record.data[Fields.STATUS] < merging_record.data[Fields.STATUS]:
            main_record.set_status(merging_record.data[Fields.STATUS])
        else:
            merging_record.set_status(main_record.data[Fields.STATUS])


def _select_author(
    record: colrev.record.record.Record, merging_record: colrev.record.record.Record
) -> str:
    if not record.has_quality_defects(
        key=Fields.AUTHOR
    ) and merging_record.has_quality_defects(key=Fields.AUTHOR):
        return record.data[Fields.AUTHOR]
    if record.has_quality_defects(
        key=Fields.AUTHOR
    ) and not merging_record.has_quality_defects(key=Fields.AUTHOR):
        return merging_record.data[Fields.AUTHOR]
    return record.data[Fields.AUTHOR]


def _select_year(
    record: colrev.record.record.Record, merging_record: colrev.record.record.Record
) -> str:
    year_pattern = r"\d{4}"
    if (
        re.match(year_pattern, merging_record.data[Fields.YEAR])
        or not re.match(
            year_pattern, record.data[Fields.YEAR]
        )  # could be "forthcoming" or "UNKNOWN"
        # record.data[Fields.YEAR] == "forthcoming"
    ):
        return merging_record.data[Fields.YEAR]
    return record.data[Fields.YEAR]


def _select_pages(
    record: colrev.record.record.Record,
    merging_record: colrev.record.record.Record,
) -> str:
    best_pages = record.data[Fields.PAGES]
    if (
        "--" in merging_record.data[Fields.PAGES]
        and "--" not in record.data[Fields.PAGES]
    ):
        best_pages = merging_record.data[Fields.PAGES]
    return best_pages


def _select_title(
    record: colrev.record.record.Record,
    merging_record: colrev.record.record.Record,
) -> str:
    default = record.data[Fields.TITLE]
    candidate = merging_record.data[Fields.TITLE]
    best_title = record.data[Fields.TITLE]

    # Note : avoid switching titles
    if default.replace(" - ", ": ") == candidate.replace(" - ", ": "):
        return default

    default_upper = colrev.env.utils.percent_upper_chars(default)
    candidate_upper = colrev.env.utils.percent_upper_chars(candidate)

    if candidate[-1] not in ["*", "1", "2"]:
        # Relatively simple rule...
        # catches cases when default is all upper or title case
        if default_upper > candidate_upper:
            best_title = candidate
    return best_title


def _select_journal(
    record: colrev.record.record.Record,
    merging_record: colrev.record.record.Record,
) -> str:
    return _select_container_title(
        record.data[Fields.JOURNAL],
        merging_record.data[Fields.JOURNAL],
    )


def _select_booktitle(
    record: colrev.record.record.Record,
    merging_record: colrev.record.record.Record,
) -> str:
    return _select_container_title(
        record.data[Fields.BOOKTITLE],
        merging_record.data[Fields.BOOKTITLE],
    )


def _select_container_title(default: str, candidate: str) -> str:
    best_journal = default

    default_upper = colrev.env.utils.percent_upper_chars(default)
    candidate_upper = colrev.env.utils.percent_upper_chars(candidate)

    # Simple heuristic to avoid abbreviations
    if "." in default and "." not in candidate:
        best_journal = candidate
    # Relatively simple rule...
    # catches cases when default is all upper or title case
    if default_upper > candidate_upper:
        best_journal = candidate
    return best_journal


CUSTOM_FIELD_SELECTORS = {
    Fields.AUTHOR: _select_author,
    Fields.PAGES: _select_pages,
    Fields.TITLE: _select_title,
    Fields.JOURNAL: _select_journal,
    Fields.BOOKTITLE: _select_booktitle,
    Fields.YEAR: _select_year,
}


def fuse_fields(
    main_record: colrev.record.record.Record,
    *,
    merging_record: colrev.record.record.Record,
    key: str,
) -> None:
    """Fuse fields based on quality heuristics"""
    # Note : the assumption is that we need masterdata_provenance notes
    # only for authors

    quality_model = colrev.record.qm.quality_model.QualityModel(
        defects_to_ignore=[
            DefectCodes.MISSING,
            DefectCodes.RECORD_NOT_IN_TOC,
            DefectCodes.INCONSISTENT_WITH_DOI_METADATA,
            DefectCodes.CONTAINER_TITLE_ABBREVIATED,
            DefectCodes.INCONSISTENT_WITH_DOI_METADATA,
        ]
    )
    quality_model.run(record=main_record)
    quality_model.run(record=merging_record)

    if key in CUSTOM_FIELD_SELECTORS:
        if key in main_record.data:
            best_value = CUSTOM_FIELD_SELECTORS[key](
                main_record,
                merging_record,
            )
            if main_record.data[key] != best_value:
                main_record.update_field(
                    key=key,
                    value=best_value,
                    source=merging_record.get_field_provenance_source(key),
                    append_edit=False,
                )
        else:
            main_record.update_field(
                key=key,
                value=merging_record.data[key],
                source=merging_record.get_field_provenance_source(key),
                append_edit=False,
            )

    elif key == Fields.FILE:
        if key in main_record.data:
            main_record.data[key] = (
                main_record.data[key] + ";" + merging_record.data.get(key, "")
            )
        else:
            main_record.data[key] = merging_record.data[key]
    elif key in [Fields.URL]:
        if (
            key in main_record.data
            and main_record.data[key].rstrip("/")
            != merging_record.data[key].rstrip("/")
            and "https" not in main_record.data[key]
        ):
            main_record.update_field(
                key=key,
                value=merging_record.data[key],
                source=merging_record.get_field_provenance_source(key),
                append_edit=False,
            )

    elif main_record.data.get(key, "") in [
        "",
        FieldValues.UNKNOWN,
    ] and merging_record.data.get(key, "") not in ["", FieldValues.UNKNOWN]:
        if _preserve_ignore_missing(main_record, key=key):
            return
        main_record.update_field(
            key=key,
            value=merging_record.data[key],
            source=merging_record.get_field_provenance_source(key),
            append_edit=False,
        )


def _preserve_ignore_missing(
    main_record: colrev.record.record.Record, *, key: str
) -> bool:
    if main_record.ignored_defect(key=key, defect=DefectCodes.MISSING):
        return True
    return False


def _merging_record_preferred(
    merging_record: colrev.record.record.Record,
    *,
    preferred_masterdata_source_prefixes: list,
) -> bool:
    if Fields.ORIGIN not in merging_record.data:
        return False
    merging_record_preferred = False
    if any(
        any(ps in origin for ps in preferred_masterdata_source_prefixes)
        for origin in merging_record.data[Fields.ORIGIN]
    ):
        merging_record_preferred = True
    return merging_record_preferred


def _prep_incoming_masterdata_merge(
    main_record: colrev.record.record.Record,
    merging_record: colrev.record.record.Record,
) -> None:
    if (
        not main_record.masterdata_is_curated()
        and merging_record.masterdata_is_curated()
    ):
        main_record.data[Fields.MD_PROV] = merging_record.data[Fields.MD_PROV]
        # Note : remove all masterdata fields
        # because the curated record may have fewer masterdata fields
        # and we iterate over the curated record (merging_record) in the next step
        for k in list(main_record.data.keys()):
            if k in FieldSet.IDENTIFYING_FIELD_KEYS and k != Fields.PAGES:
                del main_record.data[k]


def merge(
    main_record: colrev.record.record.Record,
    merging_record: colrev.record.record.Record,
    *,
    default_source: str,
    preferred_masterdata_source_prefixes: list,
) -> None:
    """General-purpose record merging
    for preparation, curated/non-curated records and records with origins


    Apply heuristics to create a fusion of the best fields based on
    quality heuristics"""

    merging_record_preferred = _merging_record_preferred(
        merging_record,
        preferred_masterdata_source_prefixes=preferred_masterdata_source_prefixes,
    )

    _merge_origins(main_record, merging_record)
    _merge_status(main_record, merging_record)

    _prep_incoming_masterdata_merge(main_record, merging_record)

    for key in list(merging_record.data.keys()):
        val, source, note = _get_merging_triple(
            merging_record, key=key, default_source=default_source
        )
        if val == "":
            continue

        # Do not change if MERGING_RECORD is not curated and main_record is curated
        if (
            main_record.masterdata_is_curated()
            and not merging_record.masterdata_is_curated()
        ):
            continue

        # Always update from curated merging_records if it is curated
        if merging_record.masterdata_is_curated():
            main_record.data[key] = merging_record.data[key]
            if key not in FieldSet.IDENTIFYING_FIELD_KEYS + [Fields.ENTRYTYPE]:
                main_record.add_field_provenance(key=key, source=source, note=note)

        # Part 1: identifying fields
        if key in FieldSet.IDENTIFYING_FIELD_KEYS:
            if merging_record_preferred:
                main_record.update_field(
                    key=key, value=str(val), source=source, append_edit=False
                )

            # Fuse best fields if none is curated
            else:
                fuse_fields(
                    main_record,
                    merging_record=merging_record,
                    key=key,
                )

        # Part 2: other fields
        else:
            # keep existing values per default
            if key in main_record.data:
                continue
            main_record.update_field(
                key=key,
                value=str(val),
                source=source,
                note=note,
                keep_source_if_equal=True,
                append_edit=False,
            )



#! /usr/bin/env python
"""The process model."""
from __future__ import annotations

import logging
import typing

import colrev.exceptions as colrev_exceptions
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.constants import RecordState

if typing.TYPE_CHECKING:  # pragma: no cover
    import colrev.process.operation


class ProcessModel:
    """The ProcessModel describes transitions between RecordStates"""

    transitions = [
        {
            "trigger": OperationsType.load,
            "source": RecordState.md_retrieved,
            "dest": RecordState.md_imported,
        },
        {
            "trigger": OperationsType.prep,
            "source": RecordState.md_imported,
            "dest": RecordState.md_needs_manual_preparation,
        },
        {
            "trigger": OperationsType.prep,
            "source": RecordState.md_imported,
            "dest": RecordState.md_prepared,
        },
        {
            "trigger": OperationsType.prep_man,
            "source": RecordState.md_needs_manual_preparation,
            "dest": RecordState.md_prepared,
        },
        {
            "trigger": OperationsType.dedupe,
            "source": RecordState.md_prepared,
            "dest": RecordState.md_processed,
        },
        {
            "trigger": OperationsType.prescreen,
            "source": RecordState.md_processed,
            "dest": RecordState.rev_prescreen_excluded,
        },
        {
            "trigger": OperationsType.prescreen,
            "source": RecordState.md_processed,
            "dest": RecordState.rev_prescreen_included,
        },
        {
            "trigger": OperationsType.pdf_get,
            "source": RecordState.rev_prescreen_included,
            "dest": RecordState.pdf_imported,
        },
        {
            "trigger": OperationsType.pdf_get,
            "source": RecordState.rev_prescreen_included,
            "dest": RecordState.pdf_needs_manual_retrieval,
        },
        {
            "trigger": OperationsType.pdf_get_man,
            "source": RecordState.pdf_needs_manual_retrieval,
            "dest": RecordState.pdf_not_available,
        },
        {
            "trigger": OperationsType.pdf_get_man,
            "source": RecordState.pdf_needs_manual_retrieval,
            "dest": RecordState.pdf_imported,
        },
        {
            "trigger": OperationsType.pdf_prep,
            "source": RecordState.pdf_imported,
            "dest": RecordState.pdf_needs_manual_preparation,
        },
        {
            "trigger": OperationsType.pdf_prep,
            "source": RecordState.pdf_imported,
            "dest": RecordState.pdf_prepared,
        },
        {
            "trigger": OperationsType.pdf_prep_man,
            "source": RecordState.pdf_needs_manual_preparation,
            "dest": RecordState.pdf_prepared,
        },
        {
            "trigger": OperationsType.screen,
            "source": RecordState.pdf_prepared,
            "dest": RecordState.rev_excluded,
        },
        {
            "trigger": OperationsType.screen,
            "source": RecordState.pdf_prepared,
            "dest": RecordState.rev_included,
        },
        {
            "trigger": OperationsType.data,
            "source": RecordState.rev_included,
            "dest": RecordState.rev_synthesized,
        },
    ]

    @classmethod
    def get_valid_transitions(cls, *, state: RecordState) -> set:
        """Get the list of valid transitions"""
        logging.getLogger("transitions").setLevel(logging.WARNING)
        return set({x["trigger"] for x in cls.transitions if x["source"] == state})

    @classmethod
    def get_preceding_states(cls, *, state: RecordState) -> set:
        """Get the states preceding the state that is given as a parameter"""

        logging.getLogger("transitions").setLevel(logging.WARNING)
        preceding_states: set[RecordState] = set()
        added = True
        while added:
            preceding_states_size = len(preceding_states)
            for transition in ProcessModel.transitions:
                if (
                    transition["dest"] in preceding_states
                    or state == transition["dest"]
                ):
                    preceding_states.add(transition["source"])  # type: ignore
            if preceding_states_size == len(preceding_states):
                added = False
        return preceding_states

    @classmethod
    def check_operation_precondition(
        cls, operation: colrev.process.operation.Operation
    ) -> None:
        """Check the preconditions for an operation"""

        def get_states_set() -> set:
            records_headers = operation.review_manager.dataset.load_records_dict(
                header_only=True
            )
            record_header_list = list(records_headers.values())

            return {el[Fields.STATUS] for el in record_header_list}

        if operation.review_manager.settings.project.delay_automated_processing:
            start_states = [
                x["source"]
                for x in ProcessModel.transitions
                if operation.type == x["trigger"]
            ]
            state: RecordState = start_states[0]  # type: ignore

            cur_state_list = get_states_set()
            # self.review_manager.logger.debug(f"cur_state_list: {cur_state_list}")
            # self.review_manager.logger.debug(f"precondition: {self.state}")
            required_absent = cls.get_preceding_states(state=state)
            # self.review_manager.logger.debug(f"required_absent: {required_absent}")
            violating_states = cur_state_list.intersection(required_absent)
            if (
                len(cur_state_list) == 0
                and not operation.type.name == "load"  # type: ignore
            ):
                raise colrev_exceptions.NoRecordsError()
            if len(violating_states) != 0:
                raise colrev_exceptions.ProcessOrderViolation(
                    operation.type.name, str(state), list(violating_states)
                )



#! /usr/bin/env python
"""Prescreen based on CLI"""
from __future__ import annotations

import textwrap

import zope.interface
from pydantic import Field

import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields


# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.package_manager.interfaces.PrescreenInterface)
class CoLRevCLIPrescreen:
    """CLI-based prescreen"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=False)

    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.prescreen_operation = prescreen_operation
        self.review_manager = prescreen_operation.review_manager

    def _print_prescreen_record(self, record: colrev.record.record.Record) -> None:
        """Print the record for prescreen operations"""

        ret_str = f"  ID: {record.data['ID']} ({record.data[Fields.ENTRYTYPE]})"
        ret_str += (
            f"\n  {Colors.GREEN}{record.data.get(Fields.TITLE, 'no title')}{Colors.END}"
            f"\n  {record.data.get(Fields.AUTHOR, 'no-author')}"
        )
        if record.data[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
            ret_str += (
                f"\n  {record.data.get(Fields.JOURNAL, 'no-journal')} "
                f"({record.data.get(Fields.YEAR, 'no-year')}) "
                f"{record.data.get(Fields.VOLUME, 'no-volume')}"
                f"({record.data.get(Fields.NUMBER, '')})"
            )
        elif record.data[Fields.ENTRYTYPE] == ENTRYTYPES.INPROCEEDINGS:
            ret_str += f"\n  {record.data.get(Fields.BOOKTITLE, 'no-booktitle')}"
        if Fields.ABSTRACT in record.data:
            lines = textwrap.wrap(
                record.data[Fields.ABSTRACT], 100, break_long_words=False
            )
            if lines:
                ret_str += f"\n  Abstract: {lines.pop(0)}\n"
                ret_str += "\n  ".join(lines) + ""

        if Fields.URL in record.data:
            ret_str += f"\n  url: {record.data[Fields.URL]}"

        if Fields.FILE in record.data:
            ret_str += f"\n  file: {record.data[Fields.FILE]}"

        print(ret_str)

    def _fun_cli_prescreen(
        self,
        *,
        prescreen_data: dict,
        split: list,
        stat_len: int,
        padding: int,
    ) -> bool:
        if self.review_manager.settings.prescreen.explanation == "":
            print(
                f"\n{Colors.ORANGE}Provide a short explanation of the prescreen{Colors.END} "
                "(why should particular papers be included?):"
            )
            print(
                'Example objective: "Include papers that focus on digital technology."'
            )
            self.review_manager.settings.prescreen.explanation = input("")
            self.review_manager.save_settings()
        else:
            print("\nIn the prescreen, the following process is followed:\n")
            print("   " + self.review_manager.settings.prescreen.explanation)
            print()

        self.review_manager.logger.debug("Start prescreen")

        if 0 == stat_len:
            self.review_manager.logger.info("No records to prescreen")

        i, quit_pressed = 0, False
        for record_dict in prescreen_data["items"]:
            if len(split) > 0:
                if record_dict[Fields.ID] not in split:
                    continue

            record = colrev.record.record.Record(record_dict)
            ret, inclusion_decision_str = "NA", "NA"
            i += 1

            print("\n\n")
            print(f"Record {i} (of {stat_len})\n")
            self._print_prescreen_record(record)

            while ret not in ["y", "n", "s", "q"]:
                ret = input(
                    "\nInclude this record "
                    "[enter y,n,s,q for yes, no, skip/decide later, quit-and-save]? "
                )
                if ret == "q":
                    quit_pressed = True
                elif ret == "s":
                    continue
                else:
                    inclusion_decision_str = ret.replace("y", "yes").replace("n", "no")

            if quit_pressed:
                self.review_manager.logger.info("Stop prescreen")
                break

            if ret == "s":
                continue

            inclusion_decision = "yes" == inclusion_decision_str
            self.prescreen_operation.prescreen(
                record=record,
                prescreen_inclusion=inclusion_decision,
                PAD=padding,
            )

        return i == stat_len

    def run_prescreen(
        self,
        records: dict,
        split: list,
    ) -> dict:
        """Prescreen records based on a cli"""

        if not split:
            split = []

        prescreen_data = self.prescreen_operation.get_data()
        stat_len = len(split) if len(split) > 0 else prescreen_data["nr_tasks"]
        padding = prescreen_data["PAD"]

        self._fun_cli_prescreen(
            prescreen_data=prescreen_data,
            split=split,
            stat_len=stat_len,
            padding=padding,
        )

        # Note : currently, it is easier to create a commit in all cases.
        # Upon continuing the prescreen, the scope-based prescreen commits the changes,
        # which is misleading.
        # Users can still squash commits.
        # Note: original: completed = self._fun_cli_prescreen(...
        # if not completed:
        #     if input("Create commit (y/n)?") != "y":
        #         return records

        self.review_manager.dataset.create_commit(
            msg="Pre-screen: manual screen (cli)", manual_author=True
        )
        return records



#! /usr/bin/env python
"""Checker for thesis-with-multiple-authors."""
from __future__ import annotations

import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class ThesisWithMultipleAuthorsChecker:
    """The ThesisWithMultipleAuthorsChecker"""

    msg = DefectCodes.THESIS_WITH_MULTIPLE_AUTHORS

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the thesis-with-multiple-authors checks"""

        if record.ignored_defect(key=Fields.AUTHOR, defect=self.msg):
            return

        if self._multiple_authored_thesis(record=record):
            record.add_field_provenance_note(key=Fields.AUTHOR, note=self.msg)
        else:
            record.remove_field_provenance_note(key=Fields.AUTHOR, note=self.msg)

    def _multiple_authored_thesis(self, *, record: colrev.record.record.Record) -> bool:
        if Fields.ENTRYTYPE not in record.data or Fields.AUTHOR not in record.data:
            return False
        if record.data["ENTRYTYPE"] in [
            "thesis",
            "phdthesis",
            "mastertsthesis",
        ] and " and " in record.data.get(Fields.AUTHOR, ""):
            return True
        return False


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(ThesisWithMultipleAuthorsChecker(quality_model))



#! /usr/bin/env python
"""Retrieving DOIs from a papers website/url as a prep operation"""
from __future__ import annotations

import collections
import re
from sqlite3 import OperationalError

import requests
import zope.interface
from pydantic import Field

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.packages.doi_org.src.doi_org as doi_connector
import colrev.record.record
import colrev.record.record_prep
import colrev.record.record_similarity
from colrev.constants import Fields

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.PrepInterface)
class DOIFromURLsPrep:
    """Prepares records by retrieving its DOI from the website (URL)"""

    settings_class = colrev.package_manager.package_settings.DefaultSettings

    ci_supported: bool = Field(default=True)

    source_correction_hint = "check with the developer"
    always_apply_changes = False

    # https://www.crossref.org/blog/dois-and-matching-regular-expressions/
    doi_regex = re.compile(r"10\.\d{4,9}/[-._;/:A-Za-z0-9]*")

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,
        settings: dict,
    ) -> None:
        self.settings = self.settings_class(**settings)
        self.prep_operation = prep_operation
        self.review_manager = prep_operation.review_manager
        self.same_record_type_required = (
            prep_operation.review_manager.settings.is_curated_masterdata_repo()
        )
        try:
            self.session = prep_operation.review_manager.get_cached_session()
        except OperationalError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="sqlite-requests-cache"
            ) from exc
        _, self.email = prep_operation.review_manager.get_committer()

    def prepare(
        self, record: colrev.record.record_prep.PrepRecord
    ) -> colrev.record.record.Record:
        """Prepare the record by retrieving its DOI from the website (url) if available"""

        if (Fields.URL not in record.data and Fields.FULLTEXT not in record.data) or (
            Fields.DOI in record.data
        ):
            return record

        try:
            url = record.data.get(Fields.URL, record.data.get(Fields.FULLTEXT, "NA"))
            headers = {"user-agent": f"{__name__}  " f"(mailto:{self.email})"}
            ret = self.session.request(
                "GET", url, headers=headers, timeout=self.prep_operation.timeout
            )
            ret.raise_for_status()
            res = re.findall(self.doi_regex, ret.text)
            if not res:
                return record

            if len(res) == 1:
                ret_dois = [(res[0], 1)]
            else:
                counter = collections.Counter(res)
                ret_dois = counter.most_common()

            if not ret_dois:
                return record

            doi, _ = ret_dois[0]

            retrieved_record_dict = {
                Fields.DOI: doi.upper(),
                Fields.ID: record.data[Fields.ID],
            }
            retrieved_record = colrev.record.record_prep.PrepRecord(
                retrieved_record_dict
            )
            doi_connector.DOIConnector.retrieve_doi_metadata(
                review_manager=self.review_manager,
                record=retrieved_record,
                timeout=self.prep_operation.timeout,
            )

            if not colrev.record.record_similarity.matches(record, retrieved_record):
                return record

            record.merge(retrieved_record, default_source=url)

        except (
            requests.exceptions.RequestException,
            colrev_exceptions.RecordNotParsableException,
        ):
            pass
        return record



#! /usr/bin/env python
"""SearchSource: CoLRev project"""
from __future__ import annotations

import shutil
import tempfile
from copy import deepcopy
from pathlib import Path

import pandas as pd
import pandasql as ps
import zope.interface
from git import Repo
from pandasql.sqldf import PandaSQLException
from pydantic import Field
from tqdm import tqdm

import colrev.exceptions as colrev_exceptions
import colrev.package_manager.interfaces
import colrev.package_manager.package_manager
import colrev.package_manager.package_settings
import colrev.record.record
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import SearchSourceHeuristicStatus
from colrev.constants import SearchType

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(colrev.package_manager.interfaces.SearchSourceInterface)
class ColrevProjectSearchSource:
    """CoLRev projects"""

    settings_class = colrev.package_manager.package_settings.DefaultSourceSettings

    source_identifier = "colrev_project_identifier"
    search_types = [SearchType.API]
    endpoint = "colrev.colrev_project"

    ci_supported: bool = Field(default=True)
    heuristic_status = SearchSourceHeuristicStatus.supported

    def __init__(
        self, *, source_operation: colrev.process.operation.Operation, settings: dict
    ) -> None:
        self.search_source = self.settings_class(**settings)
        self.review_manager = source_operation.review_manager

    # pylint: disable=colrev-missed-constant-usage
    def _validate_source(self) -> None:
        """Validate the SearchSource (parameters etc.)"""
        source = self.search_source

        self.review_manager.logger.debug(f"Validate SearchSource {source.filename}")

        if "scope" not in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "scope required in search_parameters"
            )
        if "url" not in source.search_parameters["scope"]:
            raise colrev_exceptions.InvalidQueryException(
                "url field required in search_parameters"
            )

        self.review_manager.logger.debug(f"SearchSource {source.filename} validated")

    # pylint: disable=colrev-missed-constant-usage
    @classmethod
    def add_endpoint(
        cls,
        operation: colrev.ops.search.Search,
        params: str,
    ) -> colrev.settings.SearchSource:
        """Add SearchSource as an endpoint (based on query provided to colrev search --add )"""

        # Always API search

        filename = operation.get_unique_filename(file_path_string=params.split("/")[-1])
        search_source = colrev.settings.SearchSource(
            endpoint=cls.endpoint,
            filename=filename,
            search_type=SearchType.OTHER,
            search_parameters={"scope": {"url": params}},
            comment="",
        )
        operation.add_source_and_search(search_source)
        return search_source

    def _load_records_to_import(self, *, project_url: str, project_name: str) -> dict:
        temp_path = tempfile.gettempdir() / Path(project_name)
        temp_path.mkdir()
        Repo.clone_from(project_url, temp_path, depth=1)

        try:
            project_review_manager = self.review_manager.get_connecting_review_manager(
                path_str=str(temp_path)
            )
        except colrev_exceptions.RepoSetupError as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Error retrieving records from colrev project {project_url} ({exc})"
            ) from exc

        # remote_url = project_review_manager.dataset.get_remote_url()
        # if remote_url != "NA":
        #     project_identifier = remote_url.rstrip(".git")

        project_review_manager.get_load_operation(
            notify_state_transition_operation=False,
        )
        # pylint: disable=colrev-missed-constant-usage
        self.review_manager.logger.info(
            f'Loading records from {self.search_source.search_parameters["scope"]["url"]}'
        )
        records = project_review_manager.dataset.load_records_dict()
        shutil.rmtree(temp_path)
        return records

    def _save_field_dict(self, *, input_dict: dict, input_key: str) -> list:
        list_to_return = []
        assert input_key in [Fields.MD_PROV, Fields.D_PROV]
        if input_key == Fields.MD_PROV:
            for key, value in input_dict.items():
                if isinstance(value, dict):
                    formated_node = ",".join(
                        sorted(e for e in value["note"].split(",") if "" != e)
                    )
                    list_to_return.append(f"{key}:{value['source']};{formated_node};")

        elif input_key == Fields.D_PROV:
            for key, value in input_dict.items():
                if isinstance(value, dict):
                    list_to_return.append(f"{key}:{value['source']};{value['note']};")

        return list_to_return

    def _get_stringified_record(self, *, record: dict) -> dict:
        data_copy = deepcopy(record)

        def list_to_str(*, val: list) -> str:
            return ("\n" + " " * 36).join([f.rstrip() for f in val])

        for key in [Fields.ORIGIN]:
            if key in data_copy:
                if key in [Fields.ORIGIN]:
                    data_copy[key] = sorted(list(set(data_copy[key])))
                for ind, val in enumerate(data_copy[key]):
                    if len(val) > 0:
                        if val[-1] != ";":
                            data_copy[key][ind] = val + ";"
                data_copy[key] = list_to_str(val=data_copy[key])

        for key in [Fields.MD_PROV, Fields.D_PROV]:
            if key in data_copy:
                if isinstance(data_copy[key], dict):
                    data_copy[key] = self._save_field_dict(
                        input_dict=data_copy[key], input_key=key
                    )
                if isinstance(data_copy[key], list):
                    data_copy[key] = list_to_str(val=data_copy[key])

        return data_copy

    def search(self, rerun: bool) -> None:
        """Run a search of a CoLRev project"""

        # pylint: disable=too-many-locals
        # pdf_get_operation =
        # self.review_manager.get_pdf_get_operation(notify_state_transition_operation=False)

        self._validate_source()

        colrev_project_search_feed = self.search_source.get_api_feed(
            review_manager=self.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )
        # pylint: disable=colrev-missed-constant-usage
        project_url = self.search_source.search_parameters["scope"]["url"]
        project_name = project_url.split("/")[-1].rstrip(".git")
        records_to_import = self._load_records_to_import(
            project_url=project_url, project_name=project_name
        )

        keys_to_drop = [
            Fields.MD_PROV,
            Fields.D_PROV,
            "colrev_id",
            Fields.STATUS,
            Fields.ORIGIN,
            Fields.SCREENING_CRITERIA,
            Fields.GROBID_VERSION,
        ]

        self.review_manager.logger.info("Importing selected records")
        for record_to_import in tqdm(list(records_to_import.values())):
            if "condition" in self.search_source.search_parameters["scope"]:
                res = []
                try:
                    stringified_copy = self._get_stringified_record(
                        record=record_to_import
                    )
                    stringified_copy = {k: str(v) for k, v in stringified_copy.items()}
                    # pylint: disable=possibly-unused-variable
                    rec_df = pd.DataFrame.from_records([stringified_copy])
                    query_select = "SELECT * FROM rec_df WHERE"
                    query = (
                        f"{query_select} "
                        + f"{self.search_source.search_parameters['scope']['condition']}"
                    )
                    res = ps.sqldf(query, locals())
                except PandaSQLException:
                    pass

                if len(res) == 0:
                    continue

            # Note : we need local paths for the PDFs
            # to get local_paths, we need to lookup in the registry.json
            # otherwise, we may also consider retrieving PDFs from local_index automatically
            # if Fields.FILE in record_to_import:
            #     record_to_import[Fields.FILE] = (
            #         Path(self.search_source.search_parameters["scope"][Fields.URL])
            #         / record_to_import[Fields.FILE]
            #     )

            #     pdf_get_operation.import_pdf(
            #         record=colrev.record.record.Record(record_to_import)
            #     )

            record_to_import["colrev_project_identifier"] = (
                f"{project_url}#{record_to_import['ID']}"
            )
            record_to_import = {
                k: v for k, v in record_to_import.items() if k not in keys_to_drop
            }

            try:
                colrev_project_search_feed.add_update_record(
                    retrieved_record=colrev.record.record.Record(record_to_import),
                )

            except colrev_exceptions.NotFeedIdentifiableException:
                print("not identifiable")
                continue

        colrev_project_search_feed.save()

    def prep_link_md(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.record.Record:
        """Not implemented"""
        return record

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for CoLRev projects"""

        result = {"confidence": 0.0}
        if "colrev_project" in data:
            result["confidence"] = 1.0

        return result

    def load(self, load_operation: colrev.ops.load.Load) -> dict:
        """Load the records from the SearchSource file"""

        if self.search_source.filename.suffix == ".bib":
            records = colrev.loader.load_utils.load(
                filename=self.search_source.filename,
                logger=self.review_manager.logger,
            )
            for record_id in records:
                records[record_id] = {
                    k: v
                    for k, v in records[record_id].items()
                    if k not in FieldSet.PROVENANCE_KEYS + [Fields.SCREENING_CRITERIA]
                }

            return records

        raise NotImplementedError

    def prepare(
        self, record: colrev.record.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.record.Record:
        """Source-specific preparation for CoLRev projects"""

        return record



#! /usr/bin/env python
"""Checker for record-not-in-toc."""
from __future__ import annotations

import colrev.env.local_index
import colrev.exceptions as colrev_exceptions
import colrev.record.qm.quality_model
from colrev.constants import DefectCodes
from colrev.constants import ENTRYTYPES
from colrev.constants import Fields

# pylint: disable=too-few-public-methods


class RecordNotInTOCChecker:
    """The RecordNotInTOCChecker"""

    msg = DefectCodes.RECORD_NOT_IN_TOC

    def __init__(
        self, quality_model: colrev.record.qm.quality_model.QualityModel
    ) -> None:
        self.quality_model = quality_model
        self.local_index = colrev.env.local_index.LocalIndex(verbose_mode=False)

    def run(self, *, record: colrev.record.record.Record) -> None:
        """Run the record-not-in-toc checks"""

        if record.data[Fields.ENTRYTYPE] == ENTRYTYPES.ARTICLE:
            if record.ignored_defect(key=Fields.JOURNAL, defect=self.msg):
                return
            if not self._is_in_toc(record):
                record.add_field_provenance_note(key=Fields.JOURNAL, note=self.msg)
            else:
                record.remove_field_provenance_note(key=Fields.JOURNAL, note=self.msg)
            return

        if record.data[Fields.ENTRYTYPE] == ENTRYTYPES.INPROCEEDINGS:
            if record.ignored_defect(key=Fields.BOOKTITLE, defect=self.msg):
                return
            if not self._is_in_toc(record):
                record.add_field_provenance_note(key=Fields.BOOKTITLE, note=self.msg)
            else:
                record.remove_field_provenance_note(key=Fields.BOOKTITLE, note=self.msg)
            return

    def _is_in_toc(self, record: colrev.record.record.Record) -> bool:

        try:
            self.quality_model.local_index_lock.acquire(timeout=60)
            # Search within the table-of-content in local_index
            self.local_index.retrieve_from_toc(record)
            return True

        except colrev_exceptions.RecordNotInIndexException:
            pass
        except colrev_exceptions.RecordNotInTOCException:
            return False
        finally:
            try:
                self.quality_model.local_index_lock.release()
            except ValueError:
                pass
        return True


def register(quality_model: colrev.record.qm.quality_model.QualityModel) -> None:
    """Register the checker"""
    quality_model.register_checker(RecordNotInTOCChecker(quality_model))



#! /usr/bin/env python
"""Prep for LocalIndex."""
from __future__ import annotations

import hashlib
from pathlib import Path

import colrev.env.environment_manager
import colrev.env.local_index_sqlite
import colrev.env.resources
import colrev.env.tei_parser
import colrev.exceptions as colrev_exceptions
import colrev.ops.check
import colrev.record.record
import colrev.review_manager
from colrev.constants import Fields
from colrev.constants import FieldSet
from colrev.constants import FieldValues
from colrev.constants import LocalIndexFields
from colrev.constants import RecordState

KEYS_TO_REMOVE = (
    Fields.ORIGIN,
    Fields.FULLTEXT,
    Fields.GROBID_VERSION,
    Fields.SCREENING_CRITERIA,
    Fields.METADATA_SOURCE_REPOSITORY_PATHS,
)


def _apply_status_requirements(record_dict: dict) -> None:
    if Fields.STATUS not in record_dict:
        raise colrev_exceptions.RecordNotIndexableException()

    # It is important to exclude md_prepared if the LocalIndex
    # is used to dissociate duplicates
    if record_dict[Fields.STATUS] in RecordState.get_non_processed_states():
        raise colrev_exceptions.RecordNotIndexableException()

    # Some prescreen_excluded records are not prepared
    if record_dict[Fields.STATUS] == RecordState.rev_prescreen_excluded:
        raise colrev_exceptions.RecordNotIndexableException()


def _remove_fields(record_dict: dict) -> None:
    # Do not cover deprecated fields
    for deprecated_field in ["pdf_hash"]:
        if deprecated_field in record_dict:
            print(f"Removing deprecated field: {deprecated_field}")
            del record_dict[deprecated_field]

    for field in ["note", "link"]:
        record_dict.pop(field, None)

    if Fields.SCREENING_CRITERIA in record_dict:
        del record_dict[Fields.SCREENING_CRITERIA]
    # Note: if the colrev_pdf_id has not been checked,
    # we cannot use it for retrieval or preparation.
    post_pdf_prepared_states = RecordState.get_post_x_states(
        state=RecordState.pdf_prepared
    )
    if record_dict[Fields.STATUS] not in post_pdf_prepared_states:
        record_dict.pop(Fields.PDF_ID, None)

    # Note : numbers of citations change regularly.
    # They should be retrieved from sources like crossref/doi.org
    record_dict.pop(Fields.CITED_BY, None)
    if record_dict.get(Fields.YEAR, "NA").isdigit():
        record_dict[Fields.YEAR] = int(record_dict[Fields.YEAR])
    else:
        raise colrev_exceptions.RecordNotIndexableException()

    if Fields.LANGUAGE in record_dict and len(record_dict[Fields.LANGUAGE]) != 3:
        print(f"Language not in ISO 639-3 format: {record_dict[Fields.LANGUAGE]}")
        del record_dict[Fields.LANGUAGE]


def _adjust_provenance_for_indexing(record_dict: dict) -> None:
    # Provenance should point to the original repository path.
    # If the provenance/source was example.bib (and the record is amended during indexing)
    # we wouldn't know where the example.bib belongs to.
    record = colrev.record.record.Record(record_dict)
    # Make sure that we don't add provenance information without corresponding fields
    record.align_provenance()
    for key in list(record.data.keys()):
        if not record.masterdata_is_curated():
            record.add_field_provenance(
                key=key, source=record_dict[Fields.METADATA_SOURCE_REPOSITORY_PATHS]
            )
        elif key not in FieldSet.IDENTIFYING_FIELD_KEYS + FieldSet.PROVENANCE_KEYS + [
            Fields.ID,
            Fields.ENTRYTYPE,
            Fields.METADATA_SOURCE_REPOSITORY_PATHS,
        ]:
            record.add_field_provenance(
                key=key,
                source=record_dict[Fields.METADATA_SOURCE_REPOSITORY_PATHS],
            )

    record.remove_field(key=Fields.METADATA_SOURCE_REPOSITORY_PATHS)
    record_dict = record.get_data()


def _prep_fields_for_indexing(record_dict: dict) -> None:
    # Note : file paths should be absolute when added to the LocalIndex
    if Fields.FILE in record_dict:
        pdf_path = Path(record_dict[Fields.FILE])
        if pdf_path.is_file():
            record_dict[Fields.FILE] = str(pdf_path)
        else:
            del record_dict[Fields.FILE]

    record_dict.pop(Fields.ORIGIN, None)


def _get_record_hash(record_dict: dict) -> str:
    # Note : may raise NotEnoughDataToIdentifyException
    string_to_hash = colrev.record.record.Record(record_dict).get_colrev_id()
    return hashlib.sha256(string_to_hash.encode("utf-8")).hexdigest()


def prepare_record_for_indexing(record_dict: dict) -> dict:
    """Prepare record for indexing in LocalIndex."""
    try:
        _apply_status_requirements(record_dict)
        _remove_fields(record_dict)
        _prep_fields_for_indexing(record_dict)
        _adjust_provenance_for_indexing(record_dict)
        cid_to_index = colrev.record.record.Record(record_dict).get_colrev_id()
        record_dict[Fields.COLREV_ID] = cid_to_index
        record_dict[LocalIndexFields.CITATION_KEY] = record_dict[Fields.ID]
        record_dict[LocalIndexFields.ID] = _get_record_hash(record_dict)

    except (
        colrev_exceptions.NotEnoughDataToIdentifyException
    ) as exc:  # pragma: no cover
        missing_key = ""
        if exc.missing_fields is not None:
            missing_key = ",".join(exc.missing_fields)
        raise colrev_exceptions.RecordNotIndexableException(
            missing_key=missing_key
        ) from exc

    return record_dict


def prepare_record_for_return(
    record_dict: dict,
    *,
    include_file: bool = False,
    include_colrev_ids: bool = False,
) -> colrev.record.record.Record:
    """Prepare record for return from LocalIndex."""

    for key in KEYS_TO_REMOVE:
        record_dict.pop(key, None)

    # Note: record['file'] should be an absolute path by definition
    # when stored in the LocalIndex
    if Fields.FILE in record_dict and not Path(record_dict[Fields.FILE]).is_file():
        del record_dict[Fields.FILE]

    if not include_colrev_ids and Fields.COLREV_ID in record_dict:
        del record_dict[Fields.COLREV_ID]

    if not include_file:
        colrev.record.record.Record(record_dict).remove_field(key=Fields.FILE)
        colrev.record.record.Record(record_dict).remove_field(key=Fields.PDF_ID)

    record = colrev.record.record.Record(record_dict)
    record.set_status(RecordState.md_prepared)

    if record.masterdata_is_curated():
        identifier_string = (
            record.get_field_provenance_source(FieldValues.CURATED)
            + f"#{record_dict[Fields.ID]}"
        )
        record_dict[Fields.CURATION_ID] = identifier_string

    return record



#! /usr/bin/env python
"""CoLRev screen operation: Screen documents."""
from __future__ import annotations

import math
from pathlib import Path

import colrev.exceptions as colrev_exceptions
import colrev.process.operation
import colrev.record.record
from colrev.constants import Colors
from colrev.constants import EndpointType
from colrev.constants import Fields
from colrev.constants import OperationsType
from colrev.constants import RecordState


class Screen(colrev.process.operation.Operation):
    """Screen records (based on PDFs)"""

    type = OperationsType.screen

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=self.type,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.verbose = True

    def to_screen(self, record_dict: dict) -> bool:
        """
        This method checks if the record needs to be screened.
        """
        if RecordState.pdf_prepared == record_dict[Fields.STATUS]:
            return True
        if (
            "screening_criteria" in record_dict
            and "TODO" in record_dict["screening_criteria"]
            and not RecordState.rev_excluded == record_dict[Fields.STATUS]
        ):
            return True
        return False

    def _include_all_in_screen_precondition(self, records: dict) -> bool:
        if not [
            r for r in records.values() if r[Fields.STATUS] == RecordState.pdf_prepared
        ]:
            if [
                r
                for r in records.values()
                if r[Fields.STATUS] == RecordState.md_processed
            ]:
                self.review_manager.logger.warning(
                    "No records to screen. Use "
                    f"{Colors.ORANGE}colrev prescreen --include_all{Colors.END} instead"
                )
            else:
                self.review_manager.logger.warning("No records to screen.")
            return False
        return True

    def include_all_in_screen(self, *, persist: bool) -> None:
        """Include all records in the screen"""

        if persist:
            self.review_manager.settings.screen.screen_package_endpoints = []
            self.review_manager.save_settings()

        records = self.review_manager.dataset.load_records_dict()

        if not self._include_all_in_screen_precondition(records):
            return

        selected_record_ids = [
            r[Fields.ID]
            for r in records.values()
            if RecordState.pdf_prepared == r[Fields.STATUS]
        ]

        screening_criteria = self.get_screening_criteria()

        saved_args = locals()
        saved_args["include_all"] = ""
        pad = 50
        for record_id, record_dict in records.items():
            if record_dict[Fields.STATUS] != RecordState.pdf_prepared:
                continue
            self.review_manager.report_logger.info(
                f" {record_id}".ljust(pad, " ") + "Included in screen (automatically)"
            )
            if len(screening_criteria) == 0:
                record_dict.update(screening_criteria="NA")
            else:
                record_dict.update(
                    screening_criteria=";".join([e + "=in" for e in screening_criteria])
                )
            colrev.record.record.Record(record_dict).set_status(
                RecordState.rev_included
            )

        self.review_manager.dataset.save_records_dict(records)
        self._print_stats(selected_record_ids)
        self.review_manager.dataset.create_commit(
            msg="Screen: include all",
            manual_author=False,
        )

    def get_screening_criteria(self) -> list:
        """Get the list of screening criteria from settings"""

        return list(self.review_manager.settings.screen.criteria.keys())

    def set_screening_criteria(
        self, screening_criteria: dict[str, colrev.settings.ScreenCriterion]
    ) -> None:
        """Set the screening criteria in the settings"""
        self.review_manager.settings.screen.criteria = screening_criteria
        self.review_manager.save_settings()

    def get_data(self) -> dict:
        """Get the data (records to screen)"""

        # pylint: disable=duplicate-code
        records = self.review_manager.dataset.load_records_dict()

        nr_tasks = len([r for r in records.values() if self.to_screen(r)])
        items = [r for r in records.values() if self.to_screen(r)]
        pad = 0
        if items:
            pad = min((max(len(x[Fields.ID]) for x in items) + 2), 35)
        screen_data = {"nr_tasks": nr_tasks, "PAD": pad, "items": items}

        # self.review_manager.logger.debug(
        #     self.review_manager.p_printer.pformat(screen_data)
        # )
        return screen_data

    def add_criterion(
        self, *, criterion_name: str, criterion: colrev.settings.ScreenCriterion
    ) -> None:
        """Add a screening criterion to the records and settings"""

        if criterion_name in self.review_manager.settings.screen.criteria:
            print(f"Error: criterion {criterion_name} already in settings")
            return

        self.review_manager.settings.screen.criteria[criterion_name] = criterion
        self.review_manager.save_settings()
        self.review_manager.dataset.add_setting_changes()

        records = self.review_manager.dataset.load_records_dict()
        counter = 0
        for record_dict in records.values():
            if record_dict[Fields.STATUS] not in [
                RecordState.rev_included,
                RecordState.rev_excluded,
                RecordState.rev_synthesized,
            ]:
                continue
            if Fields.SCREENING_CRITERIA not in record_dict:
                record_dict[Fields.SCREENING_CRITERIA] = ";".join(
                    f"{c}=TODO" for c in self.review_manager.settings.screen.criteria
                )
                record = colrev.record.record.Record(record_dict)
                record.set_status(RecordState.pdf_prepared)
                counter += 1
                continue
            if record_dict[Fields.STATUS] in [
                RecordState.rev_included,
                RecordState.rev_synthesized,
            ]:
                record_dict[Fields.SCREENING_CRITERIA] += f";{criterion_name}=TODO"
                # Note : we set the status to pdf_prepared because the screening
                # decisions have to be updated (resulting in inclusion or exclusion)
                record = colrev.record.record.Record(record_dict)
                record.set_status(RecordState.pdf_prepared)
                counter += 1
            if record_dict[Fields.STATUS] == RecordState.rev_excluded:
                record_dict[Fields.SCREENING_CRITERIA] += f";{criterion_name}=TODO"
                # Note : no change in colrev_status
                # because at least one of the other criteria led to exclusion decision

        print(f"Number of records that need to be screened again: {counter}")
        print()

        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(
            msg=f"Screen: add criterion: {criterion_name}",
        )
        print()

    def delete_criterion(self, criterion_to_delete: str) -> None:
        """Delete a screening criterion from the records and settings"""
        records = self.review_manager.dataset.load_records_dict()

        if criterion_to_delete in self.review_manager.settings.screen.criteria:
            del self.review_manager.settings.screen.criteria[criterion_to_delete]
            self.review_manager.save_settings()
            self.review_manager.dataset.add_setting_changes()
        else:
            print(f"Error: criterion {criterion_to_delete} not in settings")
            return

        for record_dict in records.values():
            if record_dict[Fields.STATUS] in [
                RecordState.rev_included,
                RecordState.rev_synthesized,
            ]:
                record_dict[Fields.SCREENING_CRITERIA] = (
                    record_dict[Fields.SCREENING_CRITERIA]
                    .replace(f"{criterion_to_delete}=TODO", "")
                    .replace(f"{criterion_to_delete}=in", "")
                    .replace(f"{criterion_to_delete}=out", "")
                    .replace(";;", ";")
                    .lstrip(";")
                    .rstrip(";")
                )
                # Note : colrev_status does not change
                # because the other screening criteria do not change

            if record_dict[Fields.STATUS] in [RecordState.rev_excluded]:
                record_dict[Fields.SCREENING_CRITERIA] = (
                    record_dict[Fields.SCREENING_CRITERIA]
                    .replace(f"{criterion_to_delete}=TODO", "")
                    .replace(f"{criterion_to_delete}=in", "")
                    .replace(f"{criterion_to_delete}=out", "")
                    .replace(";;", ";")
                    .lstrip(";")
                    .rstrip(";")
                )

                if (
                    "=out" not in record_dict[Fields.SCREENING_CRITERIA]
                    and "=TODO" not in record_dict[Fields.SCREENING_CRITERIA]
                ):
                    record = colrev.record.record.Record(record_dict)
                    record.set_status(RecordState.rev_included)

        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(
            msg=f"Screen: remove criterion {criterion_to_delete}",
        )

    def create_screen_split(self, *, create_split: int) -> list:
        """Split the screen between researchers"""

        data = self.get_data()
        nrecs = math.floor(data["nr_tasks"] / create_split)

        self.review_manager.logger.info(
            f"Creating screen splits for {create_split} researchers " f"({nrecs} each)"
        )

        screen_splits = []
        for _ in range(0, create_split):
            added: list[str] = []
            while len(added) < nrecs:
                added.append(next(data["items"])[Fields.ID])
            addition = "colrev screen --split " + ",".join(added)
            screen_splits.append(addition)

        return screen_splits

    def setup_custom_script(self) -> None:
        """Setup a custom screen script"""

        filedata = colrev.env.utils.get_package_file_content(
            module="colrev.ops",
            filename=Path("custom_scripts/custom_screen_script.py"),
        )

        if filedata:
            with open("custom_screen_script.py", "w", encoding="utf-8") as file:
                file.write(filedata.decode("utf-8"))

        self.review_manager.dataset.add_changes(Path("custom_screen_script.py"))

        self.review_manager.settings.screen.screen_package_endpoints.append(
            {"endpoint": "custom_screen_script"}
        )
        self.review_manager.save_settings()

    def _screen_include_all(self, records: dict) -> None:
        self.review_manager.logger.info("Screen: Include all records")
        for record_dict in records.values():
            if record_dict[Fields.STATUS] == RecordState.pdf_prepared:
                record = colrev.record.record.Record(record_dict)
                record.set_status(RecordState.rev_included)
        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(
            msg="Screen: include all",
            manual_author=False,
        )

    def _print_stats(self, selected_record_ids: list) -> None:
        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        screen_excluded = [
            r[Fields.ID]
            for r in records_headers.values()
            if RecordState.rev_excluded == r[Fields.STATUS]
            and not self.to_screen(r)
            and r[Fields.ID] in selected_record_ids
        ]
        screen_included = [
            r[Fields.ID]
            for r in records_headers.values()
            if RecordState.rev_included == r[Fields.STATUS]
            and not self.to_screen(r)
            and r[Fields.ID] in selected_record_ids
        ]

        if not screen_excluded and not screen_included:
            return

        print()
        self.review_manager.logger.info("Statistics")
        for record_dict in records_headers.values():
            if self.to_screen(record_dict):
                continue
            if record_dict[Fields.ID] in screen_excluded:
                reasons = record_dict.get(Fields.SCREENING_CRITERIA, "NA")
                if reasons == "NA":
                    reasons = ""
                else:
                    reasons = f"({reasons})"
                self.review_manager.logger.info(
                    f" {record_dict['ID']}".ljust(40)
                    + f"pdf_prepared → rev_excluded {Colors.RED}{reasons}{Colors.END}"
                )
            elif record_dict[Fields.ID] in screen_included:
                self.review_manager.logger.info(
                    f" {Colors.GREEN}{record_dict['ID']}".ljust(45)
                    + f"pdf_prepared → rev_included{Colors.END}"
                )

        nr_screen_excluded = len(screen_excluded)
        nr_screen_included = len(screen_included)

        self.review_manager.logger.info(
            "Excluded".ljust(29) + f"{nr_screen_excluded}".rjust(10, " ") + " records"
        )
        self.review_manager.logger.info(
            "Included".ljust(29) + f"{nr_screen_included}".rjust(10, " ") + " records"
        )

    def screen(
        self,
        *,
        record: colrev.record.record.Record,
        screen_inclusion: bool,
        screening_criteria: str,
        PAD: int = 40,
    ) -> None:
        """Save the screen decision"""

        PAD = 40
        if screen_inclusion:
            screening_criteria_list = self.get_screening_criteria()
            if len(screening_criteria_list) == 0:
                record.data.update(screening_criteria="NA")
            else:
                record.data.update(
                    screening_criteria=";".join(
                        [e + "=in" for e in screening_criteria_list]
                    )
                )
            record.set_status(RecordState.rev_included)

            self.review_manager.report_logger.info(
                f" {record.data['ID']}".ljust(PAD, " ") + "Included in screen"
            )
        else:
            record.data[Fields.SCREENING_CRITERIA] = screening_criteria
            record.set_status(RecordState.rev_excluded)
            self.review_manager.report_logger.info(
                f" {record.data['ID']}".ljust(PAD, " ") + "Excluded in screen"
            )

        record_dict = record.get_data()
        self.review_manager.dataset.save_records_dict(
            {record_dict[Fields.ID]: record_dict}, partial=True
        )

    def _auto_include(self, records: dict) -> list:
        selected_auto_include_ids = [
            r[Fields.ID]
            for r in records.values()
            if self.to_screen(r) and r.get("include_flag", "0") == "1"
        ]
        if not selected_auto_include_ids:
            return selected_auto_include_ids
        self.review_manager.logger.info(
            f"{Colors.GREEN}Automatically including records with include_flag{Colors.END}"
        )

        for record_dict in records.values():
            if record_dict[Fields.ID] not in selected_auto_include_ids:
                continue
            record = colrev.record.record.Record(record_dict)
            self.screen(
                record=record,
                screen_inclusion=True,
                screening_criteria="",
            )
            record.remove_field(key="include_flag")

        self.review_manager.dataset.create_commit(
            msg=f"Screen: include records {','.join(selected_auto_include_ids)}",
            manual_author=True,
        )
        return selected_auto_include_ids

    def add_abstracts_from_tei(self) -> None:
        """Add abstracts from TEI files to records without abstracts"""

        records = self.review_manager.dataset.load_records_dict()
        for record_dict in records.values():
            if (
                Fields.ABSTRACT in record_dict
                or Fields.FILE not in record_dict
                or Path(record_dict[Fields.FILE].suffix != ".pdf")
                or not Path(record_dict[Fields.FILE]).exists()
            ):
                continue

            try:
                tei = self.review_manager.get_tei(
                    pdf_path=Path(record_dict[Fields.FILE]),
                    tei_path=colrev.record.record.Record(
                        record_dict
                    ).get_tei_filename(),
                )
                record_dict[Fields.ABSTRACT] = tei.get_abstract()
            except colrev_exceptions.TEIException:
                pass
        self.review_manager.dataset.save_records_dict(records)
        self.review_manager.dataset.create_commit(msg="Add abstracts from TEI")

    @colrev.process.operation.Operation.decorate()
    def main(self, *, split_str: str = "NA") -> None:
        """Screen records for inclusion (main entrypoint)"""

        self.review_manager.logger.info("Screen")
        self.review_manager.logger.info(
            "In the screen, records are included or excluded "
            "based on the PDFs and screening criteria."
        )
        self.review_manager.logger.info(
            "See https://colrev-environment.github.io/colrev/manual/pdf_screen/screen.html"
        )

        # pylint: disable=duplicate-code
        split = []
        if split_str != "NA":
            split = split_str.split(",")
            if "" in split:
                split.remove("")

        records = self.review_manager.dataset.load_records_dict()

        package_manager = self.review_manager.get_package_manager()

        if not self.review_manager.settings.screen.screen_package_endpoints:
            self._screen_include_all(records)
            return

        for (
            screen_package_endpoint
        ) in self.review_manager.settings.screen.screen_package_endpoints:
            screen_class = package_manager.get_package_endpoint_class(
                package_type=EndpointType.screen,
                package_identifier=screen_package_endpoint["endpoint"],
            )
            endpoint = screen_class(
                screen_operation=self, settings=screen_package_endpoint
            )

            selected_auto_include_ids = self._auto_include(records)

            selected_record_ids = [
                r[Fields.ID]
                for r in records.values()
                if self.to_screen(r) and not r.get("include_flag", "0") == "1"
            ]
            if split:
                split = [x for x in selected_record_ids if x in split]

            endpoint.run_screen(records, selected_record_ids)  # type: ignore

            self._print_stats(selected_record_ids + selected_auto_include_ids)

        self.review_manager.logger.info(
            f"{Colors.GREEN}Completed screen operation{Colors.END}"
        )



#!/usr/bin/env python
"""Constants for CoLRev"""
# pylint: disable=too-few-public-methods
# pylint: disable=colrev-missed-constant-usage
import importlib.util
import typing
from enum import Enum
from enum import EnumMeta
from pathlib import Path


class Filepaths:
    """Filepaths for CoLRev"""

    # Environment-specific paths
    LOCAL_ENVIRONMENT_DIR = Path.home().joinpath(".colrev")
    CURATIONS_PATH = LOCAL_ENVIRONMENT_DIR.joinpath("curated_metadata")

    LOCAL_INDEX_SQLITE_FILE = LOCAL_ENVIRONMENT_DIR / Path("sqlite_index.db")
    TEI_INDEX_DIR = LOCAL_ENVIRONMENT_DIR / Path(".tei_index/")

    REGISTRY_FILE = LOCAL_ENVIRONMENT_DIR.joinpath(Path("registry.json"))

    PREP_REQUESTS_CACHE_FILE = LOCAL_ENVIRONMENT_DIR / Path("prep_requests_cache")

    COVERPAGES = LOCAL_ENVIRONMENT_DIR / Path(".coverpages")
    LASTPAGES = LOCAL_ENVIRONMENT_DIR / Path(".lastpages")

    colrev_spec = importlib.util.find_spec("colrev")
    if colrev_spec is not None:  # pragma: no cover
        if colrev_spec.origin is not None:  # pragma: no cover
            COLREV_PATH = Path(colrev_spec.origin).parents[1]

            # Contains the official list of packages for the documentation
            PACKAGES_JSON = COLREV_PATH / Path("colrev/package_manager/packages.json")
            PACKAGES_ENDPOINTS_JSON = COLREV_PATH / Path(
                "colrev/package_manager/package_endpoints.json"
            )


class FileSets:
    """File sets for CoLRev"""

    DEFAULT_GIT_IGNORE_ITEMS = [
        ".history",
        ".colrev",
        "__pycache__",
        "*.bib.sav",
        "venv",
        ".corrections",
        ".report.log",
        "output",
        "data/pdfs",
        "data/dedupe",
        "data/prep",
        "data/pdf_get_man/missing_pdf_files.csv",
        "data/.tei/",
        "data/prep_man/records_prep_man.bib",
        "data/data/sample_references.bib",
        ".DS_Store",
    ]
    DEPRECATED_GIT_IGNORE_ITEMS = [
        "missing_pdf_files.csv",
        "manual_cleansing_statistics.csv",
        ".references_learned_settings",
        "pdfs",
        ".tei",
        "data.csv",
        "requests_cache.sqlite",
    ]


class ENTRYTYPES:
    """Constants for record ENTRYTYPEs"""

    ARTICLE = "article"
    INPROCEEDINGS = "inproceedings"
    BOOK = "book"
    INBOOK = "inbook"
    PROCEEDINGS = "proceedings"

    INCOLLECTION = "incollection"
    PHDTHESIS = "phdthesis"
    THESIS = "thesis"
    MASTERSTHESIS = "masterthesis"
    BACHELORTHESIS = "bachelorthesis"
    TECHREPORT = "techreport"
    UNPUBLISHED = "unpublished"
    MISC = "misc"
    SOFTWARE = "software"
    ONLINE = "online"
    CONFERENCE = "conference"

    @classmethod
    def get_all(cls) -> list:
        """Get all ENTRYTYPES"""
        return [
            cls.ARTICLE,
            cls.INPROCEEDINGS,
            cls.BOOK,
            cls.INBOOK,
            cls.PROCEEDINGS,
            cls.INCOLLECTION,
            cls.PHDTHESIS,
            cls.MASTERSTHESIS,
            cls.BACHELORTHESIS,
            cls.TECHREPORT,
            cls.UNPUBLISHED,
            cls.MISC,
            cls.SOFTWARE,
            cls.ONLINE,
            cls.CONFERENCE,
        ]


class Fields:
    """Constant field names"""

    ID = "ID"
    ENTRYTYPE = "ENTRYTYPE"

    MD_PROV = "colrev_masterdata_provenance"
    D_PROV = "colrev_data_provenance"
    ORIGIN = "colrev_origin"
    STATUS = "colrev_status"
    PDF_ID = "colrev_pdf_id"

    TITLE = "title"
    AUTHOR = "author"
    YEAR = "year"
    JOURNAL = "journal"
    BOOKTITLE = "booktitle"
    CHAPTER = "chapter"
    PUBLISHER = "publisher"
    VOLUME = "volume"
    NUMBER = "number"
    PAGES = "pages"
    EDITOR = "editor"
    EDITION = "edition"
    ADDRESS = "address"
    INSTITUTION = "institution"
    MONTH = "month"
    SERIES = "series"
    SCHOOL = "school"
    LANGUAGE = "language"
    DATE = "date"
    FILE = "file"
    DOI = "doi"
    URL = "url"
    # TBD: no LINK field?
    ISSN = "issn"
    ISBN = "isbn"
    FULLTEXT = "fulltext"
    ABSTRACT = "abstract"
    KEYWORDS = "keywords"

    CONTAINER_TITLE = "container_title"

    SCREENING_CRITERIA = "screening_criteria"
    PRESCREEN_EXCLUSION = "prescreen_exclusion"
    RETRACTED = "retracted"

    # References and citations
    # First-level references (backward) and citations (forward) can be stored in the record.

    # CITED_BY refers to forward citations
    # (as extracted from GoogleScholar cited_by for example)
    # Note : rename to CITED_BY_COUNT and use CITED_BY for a list of citing papers?
    CITED_BY = "cited_by"

    # REFERENCES is a list containing a dictionary for each reference in the bibliography
    # each reference can have the same fields as constants.Fields
    # specific additional fields are TEI_ID and NR_INTEXT_CITATIONS
    REFERENCES = "references"
    TEI_ID = "tei_id"
    NR_INTEXT_CITATIONS = "nr_intext_citations"

    TEXT_FROM_PDF = "text_from_pdf"
    NR_PAGES_IN_FILE = "nr_pages_in_file"

    COLREV_ID = "colrev_id"
    CURATION_ID = "curation_ID"
    METADATA_SOURCE_REPOSITORY_PATHS = "metadata_source_repository_paths"
    GROBID_VERSION = "grobid-version"

    DBLP_KEY = "colrev.dblp.dblp_key"
    SEMANTIC_SCHOLAR_ID = "colrev.semantic_scholar.id"
    WEB_OF_SCIENCE_ID = "colrev.web_of_science.unique-id"
    PUBMED_ID = "colrev.pubmed.pubmedid"
    PMCID = "colrev.pubmed.pmcid"
    # https://www.nlm.nih.gov/bsd/mms/medlineelements.html#pmc
    EUROPE_PMC_ID = "colrev.europe_pmc.europe_pmc_id"

    GITHUB_VERSION = "colrev.github.version"
    GITHUB_LICENSE = "colrev.github.license"
    GITHUB_LANGUAGE = "colrev.github.language"
    GITHUB_NR_CONTRIBUTORS = "colrev.github.nr_contributors"
    GITHUB_NR_COMMITS = "colrev.github.nr_commits"


class LocalIndexFields:
    """Fields used in the local index"""

    ID = "id"
    CITATION_KEY = "citation_key"
    BIBTEX = "bibtex"
    TEI = "tei"
    DBLP_KEY = "dblp_key"
    TOC_KEY = "toc_key"


class FieldValues:
    """Constant field values"""

    UNKNOWN = "UNKNOWN"
    FORTHCOMING = "forthcoming"
    RETRACTED = "retracted"
    CURATED = "CURATED"
    ANONYMOUS = "Anonymous"


class FieldSet:
    """Constant field sets"""

    # """Keys of identifying fields considered for masterdata provenance"""
    PROVENANCE_KEYS = [
        Fields.MD_PROV,
        Fields.D_PROV,
        Fields.ORIGIN,
        Fields.STATUS,
        Fields.PDF_ID,
    ]

    IDENTIFYING_FIELD_KEYS = [
        Fields.TITLE,
        Fields.AUTHOR,
        Fields.YEAR,
        Fields.JOURNAL,
        Fields.BOOKTITLE,
        Fields.CHAPTER,
        Fields.PUBLISHER,
        Fields.VOLUME,
        Fields.NUMBER,
        Fields.PAGES,
        Fields.EDITOR,
        Fields.INSTITUTION,
    ]

    STANDARDIZED_FIELD_KEYS = (
        IDENTIFYING_FIELD_KEYS
        + PROVENANCE_KEYS
        + [
            Fields.ID,
            Fields.ENTRYTYPE,
            Fields.DOI,
            Fields.URL,
            Fields.ISSN,
            Fields.ISBN,
            Fields.FULLTEXT,
            Fields.ABSTRACT,
            Fields.KEYWORDS,
            Fields.CITED_BY,
            Fields.FILE,
            Fields.MONTH,
            Fields.SERIES,
            Fields.EDITION,
            Fields.LANGUAGE,
            Fields.DATE,
            Fields.SCHOOL,
            Fields.ADDRESS,
            Fields.RETRACTED,
            Fields.CURATION_ID,
            Fields.DBLP_KEY,
            Fields.SEMANTIC_SCHOLAR_ID,
            Fields.WEB_OF_SCIENCE_ID,
            Fields.PUBMED_ID,
            Fields.PMCID,
        ]
    )
    """Standardized field keys"""

    TIME_VARIANT_FIELDS = [Fields.CITED_BY]

    # Fields that are stored as lists (items separated by newlines)
    LIST_FIELDS = [Fields.ORIGIN]

    MASTERDATA = IDENTIFYING_FIELD_KEYS + [FieldValues.CURATED]
    NO_PROVENANCE = PROVENANCE_KEYS + [
        Fields.ID,
        Fields.ENTRYTYPE,
        Fields.COLREV_ID,
        Fields.METADATA_SOURCE_REPOSITORY_PATHS,
    ]


class RecordState(Enum):
    """The possible RecordStates stored in the colrev_status field
    (corresponding to the ProcessModel)"""

    # pylint: disable=invalid-name

    # without the md_retrieved state, we could not display the load transition
    md_retrieved = 1
    """Record is retrieved and stored in the ./search directory"""
    md_imported = 2
    """Record is imported into the RECORDS_FILE"""
    md_needs_manual_preparation = 3
    """Record requires manual preparation
    (colrev_masterdata_provenance provides hints)"""
    md_prepared = 4
    """Record is prepared (no missing or incomplete fields, inconsistencies checked)"""
    md_processed = 5
    """Record has been checked for duplicate associations
    with any record in RecordState md_processed or later"""
    rev_prescreen_excluded = 6
    """Record was excluded in the prescreen (based on titles/abstracts)"""
    rev_prescreen_included = 7
    """Record was included in the prescreen (based on titles/abstracts)"""
    pdf_needs_manual_retrieval = 8
    """Record marked for manual PDF retrieval"""
    pdf_imported = 9
    """PDF imported and marked for preparation"""
    pdf_not_available = 10
    """PDF is not available"""
    pdf_needs_manual_preparation = 11
    """PDF marked for manual preparation"""
    pdf_prepared = 12
    """PDF prepared"""
    rev_excluded = 13
    """Record excluded in screen (full-text)"""
    rev_included = 14
    """Record included in screen (full-text)"""
    rev_synthesized = 15
    """Record synthesized"""
    # Note : TBD: rev_coded

    def __str__(self) -> str:
        return f"{self.name}"

    def __lt__(self, other) -> bool:  # type: ignore
        if self.__class__ == RecordState and other.__class__ == RecordState:
            return self.value < other.value
        raise NotImplementedError

    @classmethod
    def get_states_requiring_file(cls) -> list:
        """Get the states that require a file"""
        return [
            RecordState.pdf_imported,
            RecordState.pdf_needs_manual_preparation,
            RecordState.pdf_prepared,
            RecordState.rev_excluded,
            RecordState.rev_included,
            RecordState.rev_synthesized,
        ]

    @classmethod
    def get_non_processed_states(cls) -> list:
        """Get the states that correspond to not-yet-processed"""
        return [
            RecordState.md_retrieved,
            RecordState.md_imported,
            RecordState.md_prepared,
            RecordState.md_needs_manual_preparation,
        ]

    @classmethod
    def get_post_x_states(cls, *, state: "RecordState") -> typing.Set["RecordState"]:
        """Get the states after state x (passed as a parameter)"""
        # pylint: disable=too-many-return-statements
        if state == RecordState.md_prepared:
            return {
                RecordState.md_prepared,
                RecordState.md_processed,
                RecordState.rev_prescreen_included,
                RecordState.rev_prescreen_excluded,
                RecordState.pdf_needs_manual_retrieval,
                RecordState.pdf_imported,
                RecordState.pdf_not_available,
                RecordState.pdf_needs_manual_preparation,
                RecordState.pdf_prepared,
                RecordState.rev_excluded,
                RecordState.rev_included,
                RecordState.rev_synthesized,
            }
        if state == RecordState.md_processed:
            return {
                RecordState.md_processed,
                RecordState.rev_prescreen_included,
                RecordState.rev_prescreen_excluded,
                RecordState.pdf_needs_manual_retrieval,
                RecordState.pdf_imported,
                RecordState.pdf_not_available,
                RecordState.pdf_needs_manual_preparation,
                RecordState.pdf_prepared,
                RecordState.rev_excluded,
                RecordState.rev_included,
                RecordState.rev_synthesized,
            }
        if state == RecordState.rev_prescreen_included:
            return {
                RecordState.rev_prescreen_included,
                RecordState.rev_prescreen_excluded,
                RecordState.pdf_needs_manual_retrieval,
                RecordState.pdf_imported,
                RecordState.pdf_not_available,
                RecordState.pdf_needs_manual_preparation,
                RecordState.pdf_prepared,
                RecordState.rev_excluded,
                RecordState.rev_included,
                RecordState.rev_synthesized,
            }
        if state == RecordState.pdf_imported:
            return {
                RecordState.pdf_imported,
                RecordState.pdf_needs_manual_preparation,
                RecordState.pdf_prepared,
                RecordState.rev_excluded,
                RecordState.rev_included,
                RecordState.rev_synthesized,
            }
        if state == RecordState.pdf_prepared:
            return {
                RecordState.pdf_prepared,
                RecordState.rev_excluded,
                RecordState.rev_included,
                RecordState.rev_synthesized,
            }
        if state == RecordState.rev_included:
            return {
                RecordState.rev_excluded,
                RecordState.rev_included,
                RecordState.rev_synthesized,
            }
        if state == RecordState.rev_synthesized:
            return {RecordState.rev_synthesized}

        raise ValueError(f"state {state}")


class DefectCodes:
    """Constant defect codes"""

    MISSING = "missing"
    RECORD_NOT_IN_TOC = "record-not-in-toc"
    INCONSISTENT_WITH_ENTRYTYPE = "inconsistent-with-entrytype"
    CONTAINER_TITLE_ABBREVIATED = "container-title-abbreviated"
    DOI_NOT_MATCHING_PATTERN = "doi-not-matching-pattern"
    ERRONEOUS_SYMBOL_IN_FIELD = "erroneous-symbol-in-field"
    ERRONEOUS_TERM_IN_FIELD = "erroneous-term-in-field"
    ERRONEOUS_TITLE_FIELD = "erroneous-title-field"
    HTML_TAGS = "html-tags"
    IDENTICAL_VALUES_BETWEEN_TITLE_AND_CONTAINER = (
        "identical-values-between-title-and-container"
    )
    INCOMPLETE_FIELD = "incomplete-field"
    INCONSISTENT_CONTENT = "inconsistent-content"
    INCONSISTENT_WITH_DOI_METADATA = "inconsistent-with-doi-metadata"
    ISBN_NOT_MATCHING_PATTERN = "isbn-not-matching-pattern"
    LANGUAGE_FORMAT_ERROR = "language-format-error"
    LANGUAGE_UNKNOWN = "language-unknown"
    MOSTLY_ALL_CAPS = "mostly-all-caps"
    NAME_ABBREVIATED = "name-abbreviated"
    NAME_FORMAT_SEPARTORS = "name-format-separators"
    NAME_FORMAT_TITLES = "name-format-titles"
    NAME_PARTICLES = "name-particles"
    PAGE_RANGE = "page-range"
    PUBMED_ID_NOT_MATCHING_PATTERN = "pubmedid-not-matching-pattern"
    THESIS_WITH_MULTIPLE_AUTHORS = "thesis-with-multiple-authors"
    YEAR_FORMAT = "year-format"


class PDFDefectCodes:
    """Constant PDF defect codes"""

    NO_TEXT_IN_PDF = "no-text-in-pdf"
    PDF_INCOMPLETE = "pdf-incomplete"
    AUTHOR_NOT_IN_PDF = "author-not-in-pdf"
    TITLE_NOT_IN_PDF = "title-not-in-pdf"
    COVERPAGE_INCLUDED = "coverpage-included"
    LAST_PAGE_APPENDED = "last-page-appended"


class OperationsType(Enum):
    """Operation types correspond to the main state transitions (see ProcessModel)"""

    # pylint: disable=invalid-name

    search = "search"
    load = "load"
    prep = "prep"
    prep_man = "prep_man"
    dedupe = "dedupe"
    prescreen = "prescreen"
    pdf_get = "pdf_get"
    pdf_get_man = "pdf_get_man"
    pdf_prep = "pdf_prep"
    pdf_prep_man = "pdf_prep_man"
    screen = "screen"
    data = "data"

    format = "format"
    check = "check"

    def __str__(self) -> str:
        return f"{self.name}"

    @classmethod
    def get_manual_extra_operations(cls) -> list:
        """Get the manual operations that require extra manual steps"""
        return [
            OperationsType.pdf_prep_man,
            OperationsType.pdf_get_man,
            OperationsType.prep_man,
        ]


class IDPattern(Enum):
    """The pattern for generating record IDs"""

    # pylint: disable=invalid-name
    first_author_year = "first_author_year"
    three_authors_year = "three_authors_year"

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_


class SortedEnumMeta(EnumMeta):
    """SortedEnumMeta"""

    def __init__(
        cls, name: str, bases: typing.Tuple[type, ...], classdict: dict
    ) -> None:
        super().__init__(name, bases, classdict)
        # Sort the members by their values
        # pylint: disable=unsubscriptable-object
        cls._member_names_ = sorted(
            cls._member_names_, key=lambda name: cls[name].value  # type: ignore
        )


class SearchType(Enum, metaclass=SortedEnumMeta):
    """Type of search source"""

    API = "API"  # Keyword-searches
    DB = "DB"  # search-results-file with search query
    TOC = "TOC"
    BACKWARD_SEARCH = "BACKWARD_SEARCH"
    FORWARD_SEARCH = "FORWARD_SEARCH"
    FILES = "FILES"  # Replaces PDFS
    OTHER = "OTHER"
    MD = "MD"

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_

    def __str__(self) -> str:
        return f"{self.name}"

    def __lt__(self, other: "SearchType") -> bool:
        if isinstance(other, SearchType):
            return self.value < other.value
        return NotImplemented


class SearchSourceHeuristicStatus(Enum):
    """Status of the SearchSource heuristic"""

    # pylint: disable=invalid-name
    na = "not_applicable"
    oni = "output_not_identifiable"
    supported = "supported"
    todo = "to_be_implemented"

    def __str__(self) -> str:
        return f"{self.name}"  # pragma: no cover


class PDFPathType(Enum):
    """Policy for handling PDFs (create symlinks or copy files)"""

    # pylint: disable=invalid-name
    symlink = "symlink"
    copy = "copy"

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_


class ScreenCriterionType(Enum):
    """Type of screening criterion"""

    # pylint: disable=invalid-name
    inclusion_criterion = "inclusion_criterion"
    exclusion_criterion = "exclusion_criterion"

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_

    def __str__(self) -> str:
        return self.name


class ShareStatReq(Enum):
    """Record status requirements for sharing"""

    # pylint: disable=invalid-name
    none = "none"
    processed = "processed"
    screened = "screened"
    completed = "completed"

    @classmethod
    def get_options(cls) -> typing.List[str]:
        """Get the options"""
        # pylint: disable=no-member
        return cls._member_names_


# pylint: disable=colrev-missed-constant-usage
class EndpointType(Enum):
    """An enum for the types of PackageEndpoints"""

    # pylint: disable=C0103
    review_type = "review_type"
    """Endpoint for review types"""
    search_source = "search_source"
    """Endpoint for search sources"""
    prep = "prep"
    """Endpoint for prep"""
    prep_man = "prep_man"
    """Endpoint for prep-man"""
    dedupe = "dedupe"
    """Endpoint for dedupe"""
    prescreen = "prescreen"
    """Endpoint for prescreen"""
    pdf_get = "pdf_get"
    """Endpoint for pdf-get"""
    pdf_get_man = "pdf_get_man"
    """Endpoint for pdf-get-man"""
    pdf_prep = "pdf_prep"
    """Endpoint for pdf-prep"""
    pdf_prep_man = "pdf_prep_man"
    """Endpoint for pdf-prep-man"""
    screen = "screen"
    """Endpoint for screen"""
    data = "data"
    """Endpoint for data"""
    na = "na"
    """Endpoint for not available"""


class ExitCodes:
    """Exit codes"""

    SUCCESS = 0
    FAIL = 1


class Colors:
    """Colors for CLI printing"""

    RED = "\033[91m"
    GREEN = "\033[92m"
    ORANGE = "\033[93m"
    BLUE = "\033[94m"
    END = "\033[0m"



#! /usr/bin/env python
"""IEEE Xplore API"""
import json
import math
import urllib.parse
import urllib.request

import colrev.record.record
from colrev.constants import Fields

# pylint: disable=invalid-name
# pylint: disable=too-many-public-methods
# pylint: disable=colrev-missed-constant-usage


class XPLORE:
    """XPLORE API class"""

    # pylint: disable=too-many-instance-attributes
    # API ENDPOINT (all non-Open Access)
    ENDPOINT = "http://ieeexploreapi.ieee.org/api/v1/search/articles"

    # Open Access Document ENDPOINT
    OPEN_ACCESS_ENDPOINT = "http://ieeexploreapi.ieee.org/api/v1/search/document/"

    # pylint: disable=colrev-missed-constant-usage
    API_FIELDS = [
        "abstract",
        "author_url",
        "accessType",
        "article_number",
        "author_order",
        "author_terms",
        "affiliation",
        "citing_paper_count",
        "conference_dates",
        "conference_location",
        "content_type",
        "doi",
        "publisher",
        "pubtype",
        "d-year",
        "end_page",
        "facet",
        "full_name",
        "html_url",
        "ieee_terms",
        "isbn",
        "issn",
        "issue",
        "pdf_url",
        "publication_year",
        "publication_title",
        "standard_number",
        "standard_status",
        "start_page",
        "title",
        "totalfound",
        "totalsearched",
        "volume",
    ]

    FIELD_MAPPING = {
        "citing_paper_count": "citations",
        "publication_year": Fields.YEAR,
        "html_url": Fields.URL,
        "pdf_url": Fields.FULLTEXT,
        "issue": Fields.NUMBER,
    }

    # array of permitted search fields for _search_field() method
    ALLOWED_SEARCH_FIELDS = [
        "abstract",
        "affiliation",
        "article_number",
        "article_title",
        "author",
        "boolean_text",
        "content_type",
        "d-au",
        "d-pubtype",
        "d-publisher",
        "d-year",
        "doi",
        "end_year",
        "facet",
        "index_terms",
        "isbn",
        "issn",
        "is_number",
        "meta_data",
        "open_access",
        "publication_number",
        "publication_title",
        "publication_year",
        "publisher",
        "querytext",
        "start_year",
        "thesaurus_terms",
    ]

    def __init__(self, *, parameters: dict, api_key: str) -> None:
        # API key
        self.apiKey = api_key

        # flag that some search criteria has been provided
        self.queryProvided = False

        # flag for Open Access, which changes ENDPOINT in use and limits results to just Open Access
        self.usingOpenAccess = False

        # flag that article number has been provided, which overrides all other search criteria
        self.usingArticleNumber = False

        # flag that a boolean method is in use
        self.usingBoolean = False

        # flag that a facet is in use
        self.usingFacet = False

        # flag that a facet has been applied, in the event that multiple facets are passed
        self.facetApplied = False

        # default of 25 results returned
        self.resultSetMax = 25

        # maximum of 200 results returned
        self.resultSetMaxCap = 200

        # records returned default to position 1 in result set
        self.startRecord = 1

        # default sort order is ascending; could also be 'desc' for descending
        self.sortOrder = "asc"

        # field name that is being used for sorting
        self.sortField = "article_title"

        # dictionary of all search parameters in use and their values
        self.parameters: dict = {}

        # dictionary of all filters in use and their values
        self.filters: dict = {}

        parameter_methods = {}
        parameter_methods["query"] = self.queryText
        parameter_methods["parameter"] = self.queryText

        for key, value in parameters.items():
            if key in parameter_methods:
                method = parameter_methods[key]
                method(value)
            else:
                self._search_field(key, value)

    def startingResult(self, start: int) -> None:
        """Set the start position in the results
        string start   Start position in the returned data"""

        self.startRecord = math.ceil(start) if (start > 0) else 1

    def maximumResults(self, maximum: int) -> None:
        """set the maximum number of results
        string maximum   Max number of results to return"""
        self.resultSetMax = math.ceil(maximum) if (maximum > 0) else 25
        self.resultSetMax = min(self.resultSetMax, self.resultSetMaxCap)

    def resultsFilter(self, filterParam: str, value: str) -> None:
        """setting a filter on results
        string filterParam   Field used for filtering
        string value    Text to filter on"""

        filterParam = filterParam.strip().lower()
        value = value.strip()

        if len(value) > 0:
            self.filters[filterParam] = value
            self.queryProvided = True

            # Standards do not have article titles, so switch to sorting by article number
            if filterParam == "content_type" and value == "Standards":
                self.resultsSorting("publication_year", "asc")

    def resultsSorting(self, field: str, order: str) -> None:
        """Setting sort order for results
        string field   Data field used for sorting
        string order   Sort order for results (ascending or descending)"""
        field = field.strip().lower()
        order = order.strip()
        self.sortField = field
        self.sortOrder = order

    def queryText(self, value: str) -> None:
        """Text to query across metadata fields, abstract and document text"""
        self._add_parameter("querytext", value)

    def articleNumber(self, value: str) -> None:
        """Article number to query"""
        self._add_parameter("article_number", value)

    def _search_field(self, field: str, value: str) -> None:
        """Shortcut method for assigning search parameters and values
        string field   Field used for searching
        string value   Text to query"""

        field = field.strip().lower()
        if field in self.ALLOWED_SEARCH_FIELDS:
            self._add_parameter(field, value)
        else:
            print("Searches against field " + field + " are not supported")

    def facetText(self, value: str) -> None:
        """Facet text to query"""
        self._add_parameter("facet", value)

    def _add_parameter(self, parameter: str, value: str) -> None:
        """Add parameter"""
        value = value.strip()

        if len(value) > 0:
            self.parameters[parameter] = value

            # viable query criteria provided
            self.queryProvided = True

            # set flags based on parameter
            if parameter == "article_number":
                self.usingArticleNumber = True

            if parameter == "boolean_text":
                self.usingBoolean = True

            if parameter in ["facet", "d-au", "d-year", "d-pubtype", "d-publisher"]:
                self.usingFacet = True

    def openAccess(self, article: str) -> None:
        """Open Access document"""
        self.usingOpenAccess = True
        self.queryProvided = True
        self.articleNumber(article)

    def _build_open_access_query(self) -> str:
        """Creates the URL for the Open Access Document API call
        return string: full URL for querying the API"""

        url = self.OPEN_ACCESS_ENDPOINT
        url += str(self.parameters["article_number"]) + "/fulltext"
        url += "?apikey=" + str(self.apiKey)
        url += "&format=json"

        return url

    def _build_query(self) -> str:
        """Creates the URL for the non-Open Access Document API call
        return string: full URL for querying the API"""

        url = self.ENDPOINT
        url += "?apikey=" + str(self.apiKey)
        url += "&format=json"
        url += "&max_records=" + str(self.resultSetMax)
        url += "&start_record=" + str(self.startRecord)
        url += "&sort_order=" + str(self.sortOrder)
        url += "&sort_field=" + str(self.sortField)

        # add in search criteria
        # article number query takes priority over all others
        if self.usingArticleNumber:
            url += "&article_number=" + str(self.parameters["article_number"])

        # boolean query
        elif self.usingBoolean:
            url += (
                "&querytext=("
                + urllib.parse.quote(self.parameters["boolean_text"])
                + ")"
            )

        else:
            for key, value in self.parameters.items():
                if self.usingFacet and self.facetApplied is False:
                    url += "&querytext=" + urllib.parse.quote(value) + "&facet=" + key
                    self.facetApplied = True

                else:
                    url += "&" + key + "=" + urllib.parse.quote(value)

        # add in filters
        for key, value in self.filters.items():
            url += "&" + key + "=" + str(value)

        return url

    def _query_api(self, url: str) -> str:
        """Creates the URL for the API call
        string url  Full URL to pass to API
        return string: Results from API"""
        with urllib.request.urlopen(url) as con:
            content = con.read()
        return content

    def _update_special_case_fields(self, *, record_dict: dict, article: dict) -> None:
        if "start_page" in article:
            record_dict[Fields.PAGES] = article.pop("start_page")
            if "end_page" in article:
                record_dict[Fields.PAGES] += "--" + article.pop("end_page")

        if "authors" in article and "authors" in article["authors"]:
            author_list = []
            for author in article["authors"]["authors"]:
                author_list.append(author["full_name"])
            record_dict[Fields.AUTHOR] = (
                colrev.record.record_prep.PrepRecord.format_author_field(
                    " and ".join(author_list)
                )
            )

        if (
            "index_terms" in article
            and "author_terms" in article["index_terms"]
            and "terms" in article["index_terms"]["author_terms"]
        ):
            record_dict[Fields.KEYWORDS] = ", ".join(
                article["index_terms"]["author_terms"]["terms"]
            )

    def _create_record_dict(self, article: dict) -> dict:
        record_dict = {Fields.ID: article["article_number"]}
        # self.review_manager.p_printer.pprint(article)

        if article["content_type"] == "Conferences":
            record_dict[Fields.ENTRYTYPE] = "inproceedings"
            if "publication_title" in article:
                record_dict[Fields.BOOKTITLE] = article.pop("publication_title")
        else:
            record_dict[Fields.ENTRYTYPE] = "article"
            if "publication_title" in article:
                record_dict[Fields.JOURNAL] = article.pop("publication_title")

        for field in self.API_FIELDS:
            if article.get(field) is None:
                continue
            record_dict[field] = str(article.get(field))

        for api_field, rec_field in self.FIELD_MAPPING.items():
            if api_field not in record_dict:
                continue
            record_dict[rec_field] = record_dict.pop(api_field)

        self._update_special_case_fields(record_dict=record_dict, article=article)

        return record_dict

    def get_records(self) -> list:
        """Calls the API to receive the records"""

        if self.usingOpenAccess is True:
            url = self._build_open_access_query()

        else:
            url = self._build_query()

        if self.queryProvided is False:
            print("No search criteria provided")

        data = self._query_api(url)
        formattedData = json.loads(data)
        if "articles" not in formattedData:
            return []
        articles = formattedData["articles"]
        records = []
        for article in articles:
            record_dict = self._create_record_dict(article)
            record = colrev.record.record.Record(record_dict)
            records.append(record)
        return records
